# -*- coding: utf-8 -*-
"""CI RetryCommand 重试上一条用户消息（Phase 8 Tier 2）

借鉴 zhikuncode RetryCommand.java（48 行）：
- /retry 命令把上一条 user 消息重新注入会话
- 支持 string 和 list[dict] 两种 content 格式
- 返回特殊标记 "injected"，由主循环将其作为新的 prompt

集成点：main.py handle_slash_command 注册 /retry。
"""
from typing import Any, Dict, List, Optional


def find_last_user_message(messages: List[Dict[str, Any]]) -> Optional[Dict[str, Any]]:
    """倒序查找最后一条 user 消息。

    Args:
        messages: 会话消息列表

    Returns:
        最后一条 role=user 的消息，无则 None
    """
    if not messages:
        return None
    for msg in reversed(messages):
        if not isinstance(msg, dict):
            continue
        if msg.get("role") == "user":
            return msg
    return None


def extract_user_text(message: Dict[str, Any]) -> str:
    """从 user 消息中提取文本内容。

    支持两种 content 格式：
    - string: 直接返回
    - list[dict]: 合并所有 type=text 的 block 的 text 字段

    Args:
        message: user 消息 dict

    Returns:
        提取的文本，无文本时返回空字符串
    """
    if not isinstance(message, dict):
        return ""
    content = message.get("content")
    if content is None:
        return ""
    if isinstance(content, str):
        return content
    if isinstance(content, list):
        parts: List[str] = []
        for block in content:
            if isinstance(block, dict) and block.get("type") == "text":
                text = block.get("text", "")
                if text:
                    parts.append(text)
        return "".join(parts)
    # 其他类型转字符串
    return str(content)


def handle_retry(session) -> str:
    """/retry 命令处理：重试上一条 user 消息。

    Args:
        session: Session 实例，需有 messages 和 add_user 方法

    Returns:
        "injected" — 已找到并注入，主循环将其作为新 prompt
        False — 无 user 消息，未注入
    """
    last_user = find_last_user_message(session.messages)
    if last_user is None:
        return False
    text = extract_user_text(last_user)
    if not text:
        return False
    session.add_user(text)
    return "injected"



# -*- coding: utf-8 -*-
"""concurrency 子包：层级化中断控制器（借鉴 zhikuncode AbortController）"""
from .abort import (
    AbortReason,
    AbortError,
    AbortSignal,
    AbortController,
)

__all__ = [
    "AbortReason",
    "AbortError",
    "AbortSignal",
    "AbortController",
]



# -*- coding: utf-8 -*-
"""AbortController 层级化中断控制器测试"""
import threading
import time
import pytest

from abyss.concurrency import (
    AbortReason,
    AbortError,
    AbortSignal,
    AbortController,
)


class TestAbortReason:
    """AbortReason 枚举测试"""

    def test_enum_values(self):
        assert AbortReason.USER_CANCEL.value == "user_cancel"
        assert AbortReason.TIMEOUT.value == "timeout"
        assert AbortReason.PARENT_ABORTED.value == "parent_aborted"
        assert AbortReason.BUDGET_EXCEEDED.value == "budget_exceeded"

    def test_all_distinct(self):
        values = [r.value for r in AbortReason]
        assert len(set(values)) == len(values)


class TestAbortError:
    """AbortError 异常测试"""

    def test_raise_with_reason(self):
        with pytest.raises(AbortError) as exc_info:
            raise AbortError(AbortReason.USER_CANCEL)
        assert exc_info.value.reason == AbortReason.USER_CANCEL

    def test_raise_with_timeout_reason(self):
        with pytest.raises(AbortError) as exc_info:
            raise AbortError(AbortReason.TIMEOUT)
        assert exc_info.value.reason == AbortReason.TIMEOUT

    def test_message_contains_reason_value(self):
        err = AbortError(AbortReason.PARENT_ABORTED)
        assert "parent_aborted" in str(err)


class TestAbortControllerBasic:
    """AbortController 基础功能测试"""

    def test_initial_state_not_aborted(self):
        ctrl = AbortController()
        assert ctrl.is_aborted() is False
        assert ctrl.reason is None

    def test_abort_sets_state(self):
        ctrl = AbortController()
        ctrl.abort(AbortReason.USER_CANCEL)
        assert ctrl.is_aborted() is True
        assert ctrl.reason == AbortReason.USER_CANCEL

    def test_abort_default_reason(self):
        ctrl = AbortController()
        ctrl.abort()
        assert ctrl.is_aborted() is True
        assert ctrl.reason == AbortReason.USER_CANCEL

    def test_abort_is_idempotent(self):
        """CAS 保证：多次 abort 只触发一次回调"""
        ctrl = AbortController()
        call_count = [0]
        ctrl.on_abort(lambda r: call_count.__setitem__(0, call_count[0] + 1))

        ctrl.abort(AbortReason.USER_CANCEL)
        ctrl.abort(AbortReason.TIMEOUT)
        ctrl.abort(AbortReason.BUDGET_EXCEEDED)

        assert ctrl.is_aborted() is True
        assert ctrl.reason == AbortReason.USER_CANCEL  # 保留首次 reason
        assert call_count[0] == 1  # 监听器只触发一次

    def test_throw_if_aborted_when_not_aborted(self):
        ctrl = AbortController()
        # 不应抛异常
        ctrl.throw_if_aborted()

    def test_throw_if_aborted_when_aborted(self):
        ctrl = AbortController()
        ctrl.abort(AbortReason.TIMEOUT)
        with pytest.raises(AbortError) as exc_info:
            ctrl.throw_if_aborted()
        assert exc_info.value.reason == AbortReason.TIMEOUT

    def test_signal_property_returns_signal(self):
        ctrl = AbortController()
        signal = ctrl.signal
        assert isinstance(signal, AbortSignal)
        assert signal.is_aborted() is False

    def test_signal_reflects_parent_state(self):
        ctrl = AbortController()
        signal = ctrl.signal
        ctrl.abort(AbortReason.USER_CANCEL)
        assert signal.is_aborted() is True

    def test_signal_throw_if_aborted(self):
        ctrl = AbortController()
        signal = ctrl.signal
        ctrl.abort(AbortReason.BUDGET_EXCEEDED)
        with pytest.raises(AbortError) as exc_info:
            signal.throw_if_aborted()
        assert exc_info.value.reason == AbortReason.BUDGET_EXCEEDED


class TestAbortControllerListeners:
    """on_abort 监听器测试"""

    def test_single_listener_called(self):
        ctrl = AbortController()
        received = []
        ctrl.on_abort(lambda r: received.append(r))
        ctrl.abort(AbortReason.USER_CANCEL)
        assert received == [AbortReason.USER_CANCEL]

    def test_multiple_listeners_called_in_order(self):
        ctrl = AbortController()
        order = []
        ctrl.on_abort(lambda r: order.append(f"first:{r.value}"))
        ctrl.on_abort(lambda r: order.append(f"second:{r.value}"))
        ctrl.abort(AbortReason.TIMEOUT)
        assert order == ["first:timeout", "second:timeout"]

    def test_listener_exception_does_not_block_others(self):
        """单个监听器异常不应阻塞其他监听器"""
        ctrl = AbortController()
        calls = []
        ctrl.on_abort(lambda r: (_ for _ in ()).throw(ValueError("boom")))
        ctrl.on_abort(lambda r: calls.append(r))
        ctrl.abort(AbortReason.USER_CANCEL)
        assert calls == [AbortReason.USER_CANCEL]

    def test_listener_registered_after_abort_not_called(self):
        """abort 之后再注册的监听器不应被调用"""
        ctrl = AbortController()
        ctrl.abort(AbortReason.USER_CANCEL)
        calls = []
        ctrl.on_abort(lambda r: calls.append(r))
        assert calls == []

    def test_on_abort_returns_handle_for_removal(self):
        """on_abort 返回可调用 handle 用于移除监听器"""
        ctrl = AbortController()
        calls = []
        handle = ctrl.on_abort(lambda r: calls.append(r))
        handle.remove()
        ctrl.abort(AbortReason.USER_CANCEL)
        assert calls == []


class TestAbortControllerHierarchy:
    """父子层级中断传播测试"""

    def test_child_inherits_parent_abort(self):
        parent = AbortController()
        child = AbortController(parent=parent)
        parent.abort(AbortReason.USER_CANCEL)
        assert child.is_aborted() is True
        assert child.reason == AbortReason.PARENT_ABORTED

    def test_child_abort_does_not_propagate_upward(self):
        """子 abort 不应影响父"""
        parent = AbortController()
        child = AbortController(parent=parent)
        child.abort(AbortReason.TIMEOUT)
        assert child.is_aborted() is True
        assert parent.is_aborted() is False

    def test_grandchild_inherits_grandparent_abort(self):
        """祖父 abort 时孙辈也应级联中断"""
        grandparent = AbortController()
        parent = AbortController(parent=grandparent)
        child = AbortController(parent=parent)
        grandparent.abort(AbortReason.BUDGET_EXCEEDED)
        assert parent.is_aborted() is True
        assert parent.reason == AbortReason.PARENT_ABORTED
        assert child.is_aborted() is True
        assert child.reason == AbortReason.PARENT_ABORTED

    def test_multiple_children_all_aborted(self):
        parent = AbortController()
        child1 = AbortController(parent=parent)
        child2 = AbortController(parent=parent)
        child3 = AbortController(parent=parent)
        parent.abort(AbortReason.USER_CANCEL)
        assert child1.is_aborted()
        assert child2.is_aborted()
        assert child3.is_aborted()

    def test_child_listener_triggered_on_parent_abort(self):
        parent = AbortController()
        child = AbortController(parent=parent)
        child_calls = []
        child.on_abort(lambda r: child_calls.append(r))
        parent.abort(AbortReason.USER_CANCEL)
        assert child_calls == [AbortReason.PARENT_ABORTED]

    def test_cascade_after_manual_child_abort(self):
        """子已手动 abort 后，父 abort 不应再次触发子的监听器"""
        parent = AbortController()
        child = AbortController(parent=parent)
        child_calls = []
        child.on_abort(lambda r: child_calls.append(r))
        child.abort(AbortReason.TIMEOUT)
        assert child_calls == [AbortReason.TIMEOUT]
        parent.abort(AbortReason.USER_CANCEL)
        # 子监听器不应被再次触发（CAS 保证幂等）
        assert child_calls == [AbortReason.TIMEOUT]
        assert child.reason == AbortReason.TIMEOUT  # 保留首次 reason

    def test_deep_hierarchy_performance(self):
        """深层级联应在合理时间内完成（无递归栈溢出）"""
        root = AbortController()
        current = root
        for _ in range(100):
            current = AbortController(parent=current)
        leaf = current

        start = time.time()
        root.abort(AbortReason.USER_CANCEL)
        elapsed = time.time() - start
        assert leaf.is_aborted() is True
        assert elapsed < 1.0  # 100 层级联应在 1s 内完成


class TestAbortControllerConcurrency:
    """多线程并发测试"""

    def test_concurrent_abort_safe(self):
        """多线程同时 abort 应安全（CAS 保证）"""
        ctrl = AbortController()
        call_count = [0]
        ctrl.on_abort(lambda r: call_count.__setitem__(0, call_count[0] + 1))

        def aborter():
            ctrl.abort(AbortReason.USER_CANCEL)

        threads = [threading.Thread(target=aborter) for _ in range(10)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert ctrl.is_aborted() is True
        assert call_count[0] == 1  # 仅一个线程成功触发回调

    def test_concurrent_abort_with_listener(self):
        ctrl = AbortController()
        received = []
        received_lock = threading.Lock()

        def listener(reason):
            with received_lock:
                received.append(reason)

        ctrl.on_abort(listener)

        threads = [
            threading.Thread(target=ctrl.abort, args=(AbortReason.USER_CANCEL,))
            for _ in range(5)
        ]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert len(received) == 1
        assert received[0] == AbortReason.USER_CANCEL

    def test_signal_checked_from_another_thread(self):
        """子线程能感知主线程的中断信号"""
        ctrl = AbortController()
        signal = ctrl.signal
        observed = []

        def worker():
            for _ in range(100):
                if signal.is_aborted():
                    observed.append(signal.reason_at_call_time if hasattr(signal, "reason_at_call_time") else ctrl.reason)
                    return
                time.sleep(0.01)

        t = threading.Thread(target=worker)
        t.start()
        time.sleep(0.05)
        ctrl.abort(AbortReason.USER_CANCEL)
        t.join(timeout=1.0)
        assert not t.is_alive()
        assert len(observed) == 1


class TestAbortSignalReadOnly:
    """AbortSignal 只读视图测试"""

    def test_signal_cannot_abort(self):
        """AbortSignal 不应暴露 abort 方法"""
        ctrl = AbortController()
        signal = ctrl.signal
        assert not hasattr(signal, "abort")

    def test_signal_shares_state_with_controller(self):
        ctrl = AbortController()
        signal = ctrl.signal
        ctrl.abort(AbortReason.USER_CANCEL)
        assert signal.is_aborted() is True

    def test_signal_throw_if_aborted_independent(self):
        ctrl = AbortController()
        signal1 = ctrl.signal
        signal2 = ctrl.signal
        ctrl.abort(AbortReason.TIMEOUT)
        with pytest.raises(AbortError):
            signal1.throw_if_aborted()
        with pytest.raises(AbortError):
            signal2.throw_if_aborted()


class TestAbortControllerUseCases:
    """实际使用场景测试"""

    def test_use_case_interrupt_long_tool(self):
        """场景：长任务执行前检查中断"""
        ctrl = AbortController()
        results = []

        def fake_long_task():
            for i in range(10):
                ctrl.throw_if_aborted()
                results.append(i)
                time.sleep(0.001)

        t = threading.Thread(target=fake_long_task)
        t.start()
        time.sleep(0.005)
        ctrl.abort(AbortReason.USER_CANCEL)
        t.join(timeout=1.0)

        assert not t.is_alive()
        assert len(results) < 10  # 应在中间被中断
        assert ctrl.is_aborted() is True

    def test_use_case_subagent_inherits_main_interrupt(self):
        """场景：子 Agent 继承主流程中断"""
        main_ctrl = AbortController()
        subagent_ctrl = AbortController(parent=main_ctrl)

        # 主流程未中断时，子 Agent 正常执行
        assert subagent_ctrl.is_aborted() is False

        # 用户按 Esc，主流程中断
        main_ctrl.abort(AbortReason.USER_CANCEL)

        # 子 Agent 立即感知
        assert subagent_ctrl.is_aborted() is True
        with pytest.raises(AbortError):
            subagent_ctrl.throw_if_aborted()

    def test_use_case_independent_subagent_abort(self):
        """场景：子 Agent 独立中断不影响主流程"""
        main_ctrl = AbortController()
        subagent_ctrl = AbortController(parent=main_ctrl)

        # 子 Agent 超时
        subagent_ctrl.abort(AbortReason.TIMEOUT)
        assert subagent_ctrl.is_aborted() is True
        # 主流程继续
        assert main_ctrl.is_aborted() is False

    def test_use_case_budget_exceeded_terminates_all(self):
        """场景：预算超限时级联终止所有子任务"""
        root = AbortController()
        worker1 = AbortController(parent=root)
        worker2 = AbortController(parent=root)
        worker1_child = AbortController(parent=worker1)

        root.abort(AbortReason.BUDGET_EXCEEDED)

        assert worker1.reason == AbortReason.PARENT_ABORTED
        assert worker2.reason == AbortReason.PARENT_ABORTED
        assert worker1_child.reason == AbortReason.PARENT_ABORTED



# -*- coding: utf-8 -*-
"""
ShellExecTool 单元测试。

重点验证：当用户通过 `Start-Process -NoNewWindow -PassThru` 启动后台进程时，
工具不能卡死等待管道 EOF。后台进程会继承 PowerShell 的 stdout 管道，
即使 PowerShell 退出，管道仍被孙进程持有 → subprocess.run 永远读不到 EOF。

复现 bug：subprocess.run(capture_output=True) 在内部调用 communicate()，
communicate() 阻塞等待 stdout/stderr 管道 EOF。Start-Process -NoNewWindow
让孙进程继承管道写端，导致 EOF 永不到来。
"""
import subprocess
import sys
import time
import os

import pytest

from abyss.tools.shell_exec import ShellExecTool


def _run_tool_in_subprocess(queue, command: str, timeout: int):
    """在子进程中执行 ShellExecTool，避免主测试进程被卡死。

    通过 multiprocessing.Queue 把耗时和结果传回主进程。
    若子进程超时未返回，主进程会 terminate 它。
    """
    # 子进程内重新 import，避免 fork 前的状态污染
    from abyss.tools.shell_exec import ShellExecTool as _Tool
    tool = _Tool()
    start = time.monotonic()
    try:
        result = tool.execute(command=command, timeout=timeout)
        queue.put((time.monotonic() - start, result))
    except Exception as exc:
        queue.put((time.monotonic() - start, {"success": False, "error": str(exc)}))


def _cleanup_ping():
    """清理可能残留的 ping.exe 进程（仅在 Windows 测试中可能被启动）。"""
    if sys.platform == "win32":
        subprocess.run(
            "taskkill /F /IM ping.exe",
            shell=True,
            capture_output=True,
        )


@pytest.mark.skipif(sys.platform != "win32", reason="Start-Process 是 Windows 特性")
def test_start_process_does_not_block_when_spawning_background_process():
    """回归测试：Start-Process -NoNewWindow -PassThru 不能让工具卡死。

    场景：用户启动后台服务（如 python -m http.server），期望工具立即返回
    Start-Process 的 PassThru 进程对象。

    Bug 行为：subprocess.run 永远等不到管道 EOF → 工具卡死。
    期望行为：PowerShell 退出后，工具在数秒内返回。
    """
    # 用 ping -t 作为后台进程：永远运行、容易清理、不依赖 python 路径
    # -NoNewWindow 让 ping 继承 PowerShell 的 stdout 管道 → 触发 bug
    command = 'Start-Process -NoNewWindow -PassThru ping -ArgumentList \'-t\',\'127.0.0.1\''

    # multiprocessing 上下文：若工具卡死，主进程可以 terminate 子进程
    import multiprocessing
    ctx = multiprocessing.get_context("spawn")
    queue = ctx.Queue()
    proc = ctx.Process(
        target=_run_tool_in_subprocess,
        args=(queue, command, 5),
    )
    proc.start()
    # 给工具 15 秒返回（5 秒 timeout + 10 秒余量，足够正常情况）
    proc.join(timeout=15)

    try:
        if proc.is_alive():
            # 工具卡死了 — 这就是 bug
            proc.terminate()
            proc.join(timeout=5)
            pytest.fail(
                "ShellExecTool 卡死：Start-Process -NoNewWindow 后工具未在 15s 内返回"
            )
        elapsed, result = queue.get(timeout=2)
    finally:
        _cleanup_ping()

    # 期望：工具在合理时间内返回（远小于 15s）
    assert elapsed < 10, (
        f"工具耗时 {elapsed:.1f}s 过长 — Start-Process 应立即返回，"
        f"实际结果: {result}"
    )
    # Start-Process 应当成功（returncode 0）
    assert result.get("success") is True, f"命令应成功，实际: {result}"
    # PassThru 的进程对象应当被捕获到 stdout 中（验证 read1() 不被
    # TextIOWrapper 预读阻塞 — 孙进程持有管道时 read() 会卡在填缓冲区）
    stdout = result.get("stdout", "")
    assert "PING" in stdout, (
        f"PassThru 进程对象应出现在 stdout 中，实际 stdout: {stdout!r}"
    )


@pytest.mark.skipif(sys.platform != "win32", reason="Start-Process 是 Windows 特性")
def test_shell_exec_normal_command_still_returns_output():
    """回归保护：修复后台进程 bug 不能破坏普通命令的输出捕获。

    普通命令（无 Start-Process）应当照常返回 stdout/stderr。
    """
    tool = ShellExecTool()
    result = tool.execute(command="Write-Output 'hello-abyss'", timeout=10)
    assert result.get("success") is True
    assert "hello-abyss" in result.get("stdout", "")


# ========== 路径安全测试（防止 shell_exec 绕过 file_read 限制）==========

class TestShellExecPathSecurity:
    """验证 shell_exec 检测读取命令的目标路径，防止绕过 PathSecurityService。

    复现 bug：file_read 拒绝读取系统文件后，AI 改用 Get-Content 命令绕过限制。
    修复：在 shell_exec.execute 中调用 _check_read_paths 检测 cat/type/Get-Content 等
    命令的目标路径，越界路径直接拒绝。
    """

    def test_get_content_system_file_rejected(self):
        """Get-Content 读取系统 hosts 应被拒绝。"""
        tool = ShellExecTool()
        result = tool.execute(
            command='Get-Content "C:\\Windows\\System32\\drivers\\etc\\hosts"',
            timeout=5,
        )
        assert result["success"] is False
        assert "shell_exec 读取路径被拒绝" in result["error"]
        assert "hosts" in result["error"]

    def test_cat_etc_passwd_rejected(self):
        """cat 读取 /etc/passwd 应被拒绝。"""
        tool = ShellExecTool()
        result = tool.execute(command="cat /etc/passwd", timeout=5)
        assert result["success"] is False
        assert "shell_exec 读取路径被拒绝" in result["error"]

    def test_type_system_file_rejected(self):
        """type 读取系统文件应被拒绝。"""
        tool = ShellExecTool()
        result = tool.execute(
            command='type "C:\\Windows\\win.ini"',
            timeout=5,
        )
        assert result["success"] is False
        assert "shell_exec 读取路径被拒绝" in result["error"]

    def test_get_item_system_file_rejected(self):
        """Get-Item 读取系统文件应被拒绝。"""
        tool = ShellExecTool()
        result = tool.execute(
            command='Get-Item "C:\\Windows\\System32\\drivers\\etc\\hosts"',
            timeout=5,
        )
        assert result["success"] is False
        assert "shell_exec 读取路径被拒绝" in result["error"]

    def test_get_content_workspace_file_allowed(self, tmp_path):
        """Get-Content 读取工作目录内文件应放行。"""
        # 创建测试文件
        f = tmp_path / "test.txt"
        f.write_text("hello\n", encoding="utf-8")

        # 设置 PathSecurityService 工作目录为 tmp_path
        from abyss.security.path_security import get_default_service
        old_wd = get_default_service()._working_dir
        get_default_service().set_working_dir(str(tmp_path))
        try:
            tool = ShellExecTool()
            result = tool.execute(
                command=f'Get-Content "{f.name}"',
                timeout=5,
            )
            # 可能成功或失败（取决于 PowerShell），但不应被路径校验拒绝
            err = result.get("error", "") or ""
            assert "shell_exec 读取路径被拒绝" not in err
        finally:
            get_default_service().set_working_dir(old_wd)

    def test_non_read_command_not_affected(self):
        """非读取命令（如 Get-Date / ls）不应被路径校验影响。"""
        tool = ShellExecTool()
        result = tool.execute(command="Get-Date", timeout=5)
        # Get-Date 应成功执行
        assert result.get("success") is True

    def test_pipe_input_not_blocked(self):
        """管道输入的 Get-Content 不应被误判（无显式路径参数）。"""
        tool = ShellExecTool()
        # 这里没有显式路径参数，不应被 _check_read_paths 拒绝
        # 即使命令本身可能失败，也不应因路径校验失败
        result = tool.execute(
            command='Write-Output "hello" | Get-Content',
            timeout=5,
        )
        err = result.get("error", "") or ""
        assert "shell_exec 读取路径被拒绝" not in err

    def test_powershell_variable_not_blocked(self):
        """PowerShell 变量参数（$variable）不应被路径校验影响。"""
        tool = ShellExecTool()
        result = tool.execute(
            command="$p = 'test.txt'; Get-Content $p",
            timeout=5,
        )
        err = result.get("error", "") or ""
        assert "shell_exec 读取路径被拒绝" not in err



# -*- coding: utf-8 -*-
"""CL CronTaskService 命令族（Phase 8 Tier 2 CL）

借鉴 zhikuncode CronTaskService：
- /cron list — 列出所有定时任务
- /cron create <interval_seconds> <command> — 创建定时任务
- /cron delete <id> — 删除指定任务
- /cron cleanup — 清理过期任务
- /cron clear — 清空所有任务

集成点：main.py handle_slash_command 注册 /cron 命令族。
CronTaskService 实例由 main.py 启动时创建并注入。
"""
from typing import Dict, List

from ..tools.cron_service import CronTaskService, DurableCronTask


def _format_task(task: DurableCronTask, now: float) -> str:
    """格式化单条任务为可读字符串。"""
    import time as _time

    def _fmt_ts(ts: float) -> str:
        if not ts:
            return "从未"
        return _time.strftime("%Y-%m-%d %H:%M:%S", _time.localtime(ts))

    next_in = max(0, int(task.next_run - now))
    interval = int(task.interval_seconds)

    # 间隔格式化
    if interval >= 86400:
        interval_str = f"{interval // 86400}天"
    elif interval >= 3600:
        interval_str = f"{interval // 3600}小时{max(0, (interval % 3600) // 60)}分"
    elif interval >= 60:
        interval_str = f"{interval // 60}分钟"
    else:
        interval_str = f"{interval}秒"

    return (
        f"  [{task.id[:8]}] 间隔={interval_str}  下次={_fmt_ts(task.next_run)}\n"
        f"    命令: {task.command}\n"
        f"    上次执行: {_fmt_ts(task.last_run or 0)}"
    )


def handle_cron(args: str, service: CronTaskService) -> str:
    """/cron 命令处理：管理定时任务。

    Args:
        args: 子命令及参数（如 "list" / "create 60 /build" / "delete abc123"）
        service: CronTaskService 实例

    Returns:
        可读的命令处理结果字符串
    """
    parts = args.split(maxsplit=1) if args else []
    if not parts:
        return _usage()

    sub = parts[0].lower()
    rest = parts[1] if len(parts) > 1 else ""

    if sub == "list":
        return _handle_list(service)
    if sub == "create":
        return _handle_create(rest, service)
    if sub == "delete":
        return _handle_delete(rest, service)
    if sub == "cleanup":
        return _handle_cleanup(service)
    if sub == "clear":
        return _handle_clear(service)

    return f"未知子命令: {sub}\n\n" + _usage()


def _usage() -> str:
    return (
        "用法:\n"
        "  /cron list                    列出所有定时任务\n"
        "  /cron create <秒> <命令>      创建定时任务（如 /cron create 60 /build）\n"
        "  /cron delete <id>             删除指定任务（id 取 list 输出的前 8 位）\n"
        "  /cron cleanup                 清理 30 天前创建的过期任务\n"
        "  /cron clear                   清空所有任务（不可恢复）"
    )


def _handle_list(service: CronTaskService) -> str:
    import time as _time

    tasks = service.list()
    if not tasks:
        return "（暂无定时任务）\n\n" + _usage()

    now = _time.time()
    lines = [f"📋 共 {len(tasks)} 条定时任务：", ""]
    for t in tasks:
        lines.append(_format_task(t, now))
    return "\n".join(lines)


def _handle_create(rest: str, service: CronTaskService) -> str:
    if not rest:
        return "用法: /cron create <间隔秒数> <命令>\n示例: /cron create 60 /build"

    parts = rest.split(maxsplit=1)
    if len(parts) < 2:
        return "用法: /cron create <间隔秒数> <命令>\n示例: /cron create 60 /build"

    try:
        interval = int(parts[0])
    except ValueError:
        return f"间隔秒数必须是整数，收到: {parts[0]}"

    if interval <= 0:
        return "间隔秒数必须 > 0"

    command = parts[1].strip()
    if not command:
        return "命令不能为空"

    task_id = service.create(interval=interval, command=command)
    return f"✓ 已创建定时任务\n  id: {task_id[:8]}\n  间隔: {interval} 秒\n  命令: {command}"


def _handle_delete(rest: str, service: CronTaskService) -> str:
    if not rest:
        return "用法: /cron delete <id>（id 取 list 输出的前 8 位）"

    target = rest.strip()
    # 支持 8 位短 id 匹配
    tasks = service.list()
    matched = [t for t in tasks if t.id.startswith(target)]
    if not matched:
        return f"未找到 id 以 {target} 开头的任务"
    if len(matched) > 1:
        return f"多个任务匹配 {target}，请提供更长的前缀"

    if service.delete(matched[0].id):
        return f"✓ 已删除任务 {matched[0].id[:8]}"
    return f"删除失败：任务 {matched[0].id[:8]} 不存在"


def _handle_cleanup(service: CronTaskService) -> str:
    purged = service.cleanup()
    remaining = service.count()
    return f"✓ 清理了 {purged} 条过期任务，剩余 {remaining} 条"


def _handle_clear(service: CronTaskService) -> str:
    tasks = service.list()
    for t in tasks:
        service.delete(t.id)
    return f"✓ 已清空所有 {len(tasks)} 条任务"


# /cron 命令注册元数据（供 main.py 使用）
CRON_COMMANDS = {
    "cron": "管理定时任务（list/create/delete/cleanup/clear）",
}



# -*- coding: utf-8 -*-
"""PathSecurityService 路径安全服务（Phase 6 Tier 1-5 AP）。

借鉴 zhikuncode-main PathSecurityService.java：
- 5 重校验：符号链接、路径穿越、工作目录边界、受保护路径、Windows 绕过
- RiskLevel 3 级分类（LOW/MEDIUM/HIGH）
- needs_confirmation 标志位（受保护路径读操作）
- check_dangerous_removal 命令级拦截（rm -rf /）

与 ToolSafetyGuard 互补：
- PathSecurityService：file_write/file_edit/file_read 专用，路径校验
- ToolSafetyGuard：shell_exec bash 命令检测 + 环境变量清理

集成点：
1. `tools/file_write.py` / `file_edit.py`：写前 check_write
2. `tools/file_read.py`：读前 check_read
3. `tools/shell_exec.py`：rm 类命令前 check_dangerous_removal
"""
from __future__ import annotations

import os
import re
import sys
import threading
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Optional

__all__ = [
    "PathSecurityService",
    "PathCheckResult",
    "RiskLevel",
    "get_default_service",
]


def _find_protected_name(path: str) -> Optional[str]:
    """延迟导入 find_protected_name，避免循环导入。

    path_security -> tools.base -> tools.__init__ -> tools.file_read -> path_security
    """
    from abyss.tools.base import find_protected_name
    return find_protected_name(path)


class RiskLevel(Enum):
    """风险等级。"""

    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


@dataclass
class PathCheckResult:
    """路径校验结果。"""

    allowed: bool = False
    reason: str = ""
    normalized_path: str = ""
    risk_level: RiskLevel = RiskLevel.LOW
    needs_confirmation: bool = False

    @classmethod
    def allowed(cls, path: str, risk: RiskLevel = RiskLevel.LOW) -> "PathCheckResult":
        return cls(
            allowed=True,
            reason="ok",
            normalized_path=path,
            risk_level=risk,
            needs_confirmation=False,
        )

    @classmethod
    def denied(cls, reason: str, path: str = "", risk: RiskLevel = RiskLevel.HIGH) -> "PathCheckResult":
        return cls(
            allowed=False,
            reason=reason,
            normalized_path=path,
            risk_level=risk,
            needs_confirmation=False,
        )

    @classmethod
    def confirmation_required(cls, reason: str, path: str) -> "PathCheckResult":
        return cls(
            allowed=False,
            reason=reason,
            normalized_path=path,
            risk_level=RiskLevel.HIGH,
            needs_confirmation=True,
        )


# Windows DOS 设备名
_DOS_DEVICE_NAMES = {"CON", "PRN", "AUX", "NUL"} | {f"COM{i}" for i in range(1, 10)} | {f"LPT{i}" for i in range(1, 10)}

# Unix 设备文件路径前缀
_UNIX_DEVICE_PREFIXES = ("/dev/", "/proc/", "/sys/")

# 危险删除命令正则（rm -rf / 系列）
_DANGEROUS_REMOVAL_PATTERNS = [
    # rm -rf / 根目录（绝对路径）
    re.compile(r"\brm\s+(-\w*\s+)*-?\w*\s*rf?\s+(?:--no-preserve-root\s+)?(/|/\*|~|/\*|/\.\*)\s*$"),
    re.compile(r"\brm\s+-\w*r\w*\s+(?:--no-preserve-root\s+)?(/|~|/\*|/.+)\s*$"),
    re.compile(r"\bsudo\s+rm\s+-\w*r\w*\s+(?:--no-preserve-root\s+)?(/|~|/\*)\s*$"),
    # rm -rf /* 等通配根
    re.compile(r"\brm\s+-\w*r\w*\s+/\*"),
    # rm -rf /etc /home /var 等系统目录
    re.compile(r"\brm\s+-\w*r\w*\s+/(etc|home|var|usr|root|boot|bin|sbin|lib|opt)\b"),
]

# Windows ADS 标记
_NTFS_ADS_PATTERN = re.compile(r":[^\\/]+\b(?!$)")  # file.txt:stream

# 8.3 短文件名
_SHORT_NAME_PATTERN = re.compile(r"^[A-Za-z0-9_~]{1,8}~\d+\.[A-Za-z0-9]{1,3}$", re.IGNORECASE)


class PathSecurityService:
    """统一路径安全验证服务。

    5 重校验链：
    1. 输入合法性（None / 空 / null 字节）
    2. Windows 路径绕过（NTFS ADS / 8.3 短名 / DOS 设备名）
    3. 路径穿越（.. 序列）
    4. 工作目录边界（realpath 解析后）
    5. 受保护路径黑名单 + 符号链接逃逸
    """

    def __init__(self, working_dir: Optional[str] = None) -> None:
        self._working_dir: Optional[str] = working_dir
        self._lock = threading.Lock()

    def set_working_dir(self, path: str) -> None:
        """更新工作目录（运行时切换 workspace）。"""
        with self._lock:
            self._working_dir = path

    # ------------------------------------------------------------------ public API

    def check_read(self, path: Optional[str]) -> PathCheckResult:
        """校验读取路径。受保护路径读取需用户确认。"""
        result = self._basic_check(path)
        if result is not None:
            return result
        assert path is not None

        # 受保护路径优先检测（基于原始 path 字符串，无需 resolve）
        # 这样相对路径 ".git/config" 也能被识别
        protected = _find_protected_name(path)
        if protected is not None:
            return PathCheckResult.confirmation_required(
                f"受保护路径禁止读取: {protected}", path
            )

        # 路径穿越检测
        traversal = self._check_traversal(path)
        if traversal is not None:
            return traversal

        # 规范化 + 边界检测
        resolved = self._resolve(path)
        if resolved is None:
            return PathCheckResult.denied("cannot resolve path", path)

        # 工作目录边界
        boundary = self._check_workspace_boundary(resolved)
        if boundary is not None:
            return boundary

        # 符号链接逃逸
        symlink = self._check_symlink_escape(path, resolved)
        if symlink is not None:
            return symlink

        return PathCheckResult.allowed(resolved, RiskLevel.LOW)

    def check_write(self, path: Optional[str]) -> PathCheckResult:
        """校验写入路径。受保护路径写操作直接拒绝。"""
        result = self._basic_check(path)
        if result is not None:
            return result
        assert path is not None

        # 受保护路径优先检测（基于原始 path 字符串，无需 resolve）
        protected = _find_protected_name(path)
        if protected is not None:
            return PathCheckResult.denied(
                f"受保护路径禁止写入: {protected}", path
            )

        # Windows 绕过
        win = self._check_windows_bypass(path)
        if win is not None:
            return win

        # 路径穿越
        traversal = self._check_traversal(path)
        if traversal is not None:
            return traversal

        # 规范化 + 边界
        resolved = self._resolve(path)
        if resolved is None:
            return PathCheckResult.denied("cannot resolve path", path)

        # 工作目录边界
        boundary = self._check_workspace_boundary(resolved)
        if boundary is not None:
            return boundary

        # Unix 设备文件
        if sys.platform != "win32":
            for prefix in _UNIX_DEVICE_PREFIXES:
                if resolved.startswith(prefix):
                    return PathCheckResult.denied(f"device path blocked: {resolved}", resolved)

        # 符号链接逃逸
        symlink = self._check_symlink_escape(path, resolved)
        if symlink is not None:
            return symlink

        return PathCheckResult.allowed(resolved, RiskLevel.LOW)

    def check_dangerous_removal(self, command: str) -> Optional[str]:
        """检测危险删除命令。命中返回原因，否则返回 None。"""
        if not command:
            return None
        for pattern in _DANGEROUS_REMOVAL_PATTERNS:
            if pattern.search(command):
                return f"dangerous removal: {command.strip()[:80]}"
        return None

    # ------------------------------------------------------------------ 内部校验

    def _basic_check(self, path: Optional[str]) -> Optional[PathCheckResult]:
        """输入合法性基础检测。"""
        if path is None:
            return PathCheckResult.denied("path is None")
        if not path:
            return PathCheckResult.denied("empty path")
        if "\x00" in path:
            return PathCheckResult.denied("path contains null byte")
        return None

    def _check_windows_bypass(self, path: str) -> Optional[PathCheckResult]:
        """Windows 路径绕过检测：NTFS ADS / 8.3 短名 / DOS 设备名。"""
        if sys.platform != "win32":
            return None

        # NTFS ADS: file.txt:stream
        # 仅检查文件名部分（不含盘符）
        basename = os.path.basename(path)
        # 排除 Windows 盘符冒号（C:）
        if ":" in basename:
            # 检查是否为 ADS（basename 中含冒号且不是盘符）
            # 例如 "file.txt:hidden"
            parts = basename.split(":", 1)
            if len(parts) == 2 and parts[1] and parts[0]:
                return PathCheckResult.denied(
                    f"NTFS ADS blocked: {basename}", path
                )

        # 8.3 短文件名：FILENA~1.TXT
        if _SHORT_NAME_PATTERN.match(basename):
            return PathCheckResult.denied(
                f"8.3 short filename blocked: {basename}", path
            )

        # DOS 设备名：CON/PRN/NUL/COM1-9/LPT1-9
        name_no_ext = basename.split(".")[0].upper()
        if name_no_ext in _DOS_DEVICE_NAMES:
            return PathCheckResult.denied(
                f"DOS device name blocked: {basename}", path
            )

        return None

    def _check_traversal(self, path: str) -> Optional[PathCheckResult]:
        """路径穿越检测：.. 序列。"""
        # 拆分路径组件
        parts = re.split(r"[\\/]", path)
        if ".." in parts:
            # 进一步规范化判断：若解析后仍在工作目录内，放行
            resolved = self._resolve(path)
            if resolved is None:
                return PathCheckResult.denied(
                    f"path traversal cannot resolve: {path}", path
                )
            if not self._is_in_working_dir(resolved):
                return PathCheckResult.denied(
                    f"path traversal outside workspace: {path}", resolved
                )
        return None

    def _resolve(self, path: str) -> Optional[str]:
        """解析路径为绝对路径（含符号链接解析）。

        相对路径基于 working_dir 解析；working_dir=None 时基于进程 CWD。
        """
        try:
            p = Path(path)
            if not p.is_absolute():
                base = self._working_dir if self._working_dir else os.getcwd()
                p = Path(base) / p
            # resolve(strict=False) 解析符号链接但允许路径不存在
            return str(p.resolve(strict=False))
        except (OSError, ValueError, RuntimeError):
            return None

    def _is_in_working_dir(self, resolved: str) -> bool:
        """检查解析后路径是否在工作目录内。

        working_dir=None 时回退到进程当前工作目录（os.getcwd()）。
        这样默认就有边界保护，而不是放行所有路径（安全关键修复）。
        """
        wd = self._working_dir if self._working_dir is not None else os.getcwd()
        try:
            wd_resolved = Path(wd).resolve(strict=False)
            Path(resolved).relative_to(wd_resolved)
            return True
        except (ValueError, RuntimeError, OSError):
            return False

    def _check_workspace_boundary(self, resolved: str) -> Optional[PathCheckResult]:
        """工作目录边界检测。"""
        if not self._is_in_working_dir(resolved):
            return PathCheckResult.denied(
                f"path outside workspace boundary: {resolved}", resolved
            )
        return None

    def _check_symlink_escape(self, original: str, resolved: str) -> Optional[PathCheckResult]:
        """符号链接逃逸检测：符号链接指向工作目录外。"""
        try:
            p = Path(original)
            if not p.is_symlink():
                return None
            # 读取符号链接目标
            target = os.readlink(original) if p.exists() or os.path.islink(original) else None
            if target is None:
                return None
            # 解析目标为绝对路径
            target_path = Path(target)
            if not target_path.is_absolute():
                # 相对符号链接基于链接所在目录解析
                target_path = Path(original).parent / target_path
            target_resolved = str(target_path.resolve(strict=False))
            # 目标在工作目录外：需确认
            if not self._is_in_working_dir(target_resolved):
                return PathCheckResult.confirmation_required(
                    f"symlink escapes workspace: {original} -> {target_resolved}",
                    resolved,
                )
        except (OSError, ValueError, RuntimeError):
            # 符号链接读取失败不阻断（让其他校验处理）
            return None
        return None


# ------------------------------------------------------------------ 模块级单例

_default_service: Optional[PathSecurityService] = None
_default_lock = threading.Lock()


def get_default_service() -> PathSecurityService:
    """返回模块级默认 PathSecurityService 单例。"""
    global _default_service
    if _default_service is None:
        with _default_lock:
            if _default_service is None:
                _default_service = PathSecurityService()
    return _default_service



# -*- coding: utf-8 -*-
"""
413 两阶段恢复测试（红-绿-重构中的红阶段）。

借鉴 zhikuncode-main 的 CollapseDrain + ReactiveCompact 机制：当 LLM API
返回 HTTP 413（Payload Too Large）时，调用 on_413_callback 执行两阶段
激进压缩后重试，避免长对话直接失败。

测试覆盖：
1. 413 触发 Phase 1：回调返回 True，重试成功
2. Phase 1 后成功：日志记录「413 恢复成功（Phase 1）」
3. Phase 2 后成功：Phase 1 仍 413，Phase 2 压缩后重试成功
4. 两阶段均失败：抛 RuntimeError("413 两阶段恢复失败")
5. 非 413（429/500/503）走原有重试，不调用 on_413_callback
6. 回调返回 False（压缩失败）直接抛异常
"""
from unittest.mock import MagicMock, patch

import httpx
import pytest
from openai import APIStatusError, RateLimitError, InternalServerError

from abyss.api_client import DeepSeekClient


def _make_413_error(msg: str = "Payload Too Large") -> APIStatusError:
    """构造 status_code=413 的 APIStatusError。

    APIStatusError.__init__(message, *, response, body)，
    status_code 属性由 response.status_code 设置。
    """
    response = httpx.Response(
        status_code=413,
        request=httpx.Request("POST", "https://api.deepseek.com/chat"),
    )
    return APIStatusError(msg, response=response, body=None)


def _make_chunk(content=None, reasoning=None, finish=None):
    """构造与 _parse_chunk 兼容的最小流式 chunk。"""
    delta = MagicMock()
    delta.content = content
    delta.reasoning_content = reasoning
    delta.tool_calls = None
    choice = MagicMock()
    choice.delta = delta
    choice.finish_reason = finish
    chunk = MagicMock()
    chunk.choices = [choice]
    chunk.usage = None
    return chunk


def _make_rate_limit_error() -> RateLimitError:
    """构造 429 RateLimitError。"""
    response = httpx.Response(
        status_code=429,
        request=httpx.Request("POST", "https://api.deepseek.com/chat"),
    )
    return RateLimitError("rate limited", response=response, body=None)


def _make_500_error() -> InternalServerError:
    """构造 500 InternalServerError。"""
    response = httpx.Response(
        status_code=500,
        request=httpx.Request("POST", "https://api.deepseek.com/chat"),
    )
    return InternalServerError("server error", response=response, body=None)


def _make_503_error() -> APIStatusError:
    """构造 503 APIStatusError。"""
    response = httpx.Response(
        status_code=503,
        request=httpx.Request("POST", "https://api.deepseek.com/chat"),
    )
    return APIStatusError("unavailable", response=response, body=None)


class Test413Phase1Recovery:
    """413 触发 Phase 1 压缩后重试成功。"""

    def test_413_triggers_phase1_callback_then_succeeds(self):
        """413 → 调用 on_413_callback(1) → 回调返回 True → 重试成功。"""
        client = DeepSeekClient(api_key="fake-key")
        client.client = MagicMock()
        # 第一次抛 413，第二次返回成功流
        success_stream = iter([_make_chunk(content="hi"), _make_chunk(finish="stop")])
        client.client.chat.completions.create.side_effect = [
            _make_413_error(),
            success_stream,
        ]

        callback = MagicMock(return_value=True)
        with patch("abyss.api_client.logger") as mock_logger:
            mock_logger.api_request = MagicMock()
            mock_logger.api_retry = MagicMock()
            chunks = list(client.chat(
                messages=[{"role": "user", "content": "hi"}],
                stream=True,
                on_413_callback=callback,
            ))

        # 回调被调用一次，phase=1
        callback.assert_called_once_with(1)
        # 重试成功，拿到 chunks
        assert len(chunks) == 2
        assert chunks[0]["content"] == "hi"

    def test_phase1_success_logs_recovery_message(self):
        """Phase 1 成功后应记录「413 恢复成功（Phase 1）」日志。"""
        client = DeepSeekClient(api_key="fake-key")
        client.client = MagicMock()
        success_stream = iter([_make_chunk(content="ok"), _make_chunk(finish="stop")])
        client.client.chat.completions.create.side_effect = [
            _make_413_error(),
            success_stream,
        ]

        callback = MagicMock(return_value=True)
        with patch("abyss.api_client.logger") as mock_logger:
            mock_logger.api_request = MagicMock()
            mock_logger.api_retry = MagicMock()
            list(client.chat(
                messages=[{"role": "user", "content": "hi"}],
                stream=True,
                on_413_callback=callback,
            ))
            # 验证有日志包含「413 恢复成功」和「Phase 1」
            all_calls = []
            for c in mock_logger.api_retry.call_args_list:
                all_calls.append(str(c))
            # 至少有一条日志提到 413 恢复成功 Phase 1
            joined = " | ".join(all_calls)
            # 也检查 api_error / 其他调用
            for method_name in ("api_error", "info", "warning"):
                m = getattr(mock_logger, method_name, None)
                if m is not None and m.called:
                    for c in m.call_args_list:
                        all_calls.append(str(c))
            joined = " | ".join(all_calls)
            assert "413" in joined, f"日志应包含 413 字样，实际：{joined!r}"
            assert "Phase 1" in joined or "phase 1" in joined.lower(), (
                f"日志应包含 Phase 1 字样，实际：{joined!r}"
            )


class Test413Phase2Recovery:
    """Phase 1 仍 413，Phase 2 压缩后重试成功。"""

    def test_phase2_recovery_succeeds_after_phase1_still_413(self):
        """Phase 1 后仍 413 → 调用 on_413_callback(2) → Phase 2 后重试成功。"""
        client = DeepSeekClient(api_key="fake-key")
        client.client = MagicMock()
        success_stream = iter([_make_chunk(content="ok"), _make_chunk(finish="stop")])
        client.client.chat.completions.create.side_effect = [
            _make_413_error(),  # 首次请求
            _make_413_error(),  # Phase 1 重试
            success_stream,     # Phase 2 重试成功
        ]

        callback = MagicMock(return_value=True)
        with patch("abyss.api_client.logger") as mock_logger:
            mock_logger.api_request = MagicMock()
            mock_logger.api_retry = MagicMock()
            chunks = list(client.chat(
                messages=[{"role": "user", "content": "hi"}],
                stream=True,
                on_413_callback=callback,
            ))

        # 回调依次以 phase=1 和 phase=2 被调用
        assert callback.call_args_list == [
            ((1,), {}),
            ((2,), {}),
        ], f"回调调用顺序错误：{callback.call_args_list}"
        assert len(chunks) == 2
        assert chunks[0]["content"] == "ok"


class Test413BothPhasesFail:
    """两阶段压缩后仍 413 → 抛 RuntimeError。"""

    def test_both_phases_fail_raises_runtime_error(self):
        """Phase 1 + Phase 2 后仍 413 → 抛 RuntimeError，消息含「413 两阶段恢复失败」。"""
        client = DeepSeekClient(api_key="fake-key")
        client.client = MagicMock()
        client.client.chat.completions.create.side_effect = [
            _make_413_error(),  # 首次
            _make_413_error(),  # Phase 1 重试
            _make_413_error(),  # Phase 2 重试
        ]

        callback = MagicMock(return_value=True)
        with patch("abyss.api_client.logger") as mock_logger:
            mock_logger.api_request = MagicMock()
            mock_logger.api_retry = MagicMock()
            mock_logger.api_error = MagicMock()
            with pytest.raises(RuntimeError) as exc_info:
                list(client.chat(
                    messages=[{"role": "user", "content": "hi"}],
                    stream=True,
                    on_413_callback=callback,
                ))
        # 异常消息应包含「413 两阶段恢复失败」
        assert "413" in str(exc_info.value), (
            f"异常消息应包含 413，实际：{exc_info.value!r}"
        )
        # 回调被调用两次：phase=1, phase=2
        assert callback.call_args_list == [((1,), {}), ((2,), {})]


class TestNon413NotTriggerCallback:
    """非 413 错误码不应触发 on_413_callback。"""

    def test_429_does_not_trigger_413_callback(self):
        """429 RateLimitError 走原有重试，不调用 on_413_callback。"""
        client = DeepSeekClient(api_key="fake-key")
        client.client = MagicMock()
        client.retry_delay = 0.001
        client.client.chat.completions.create.side_effect = _make_rate_limit_error()

        callback = MagicMock(return_value=True)
        with patch("abyss.api_client.logger") as mock_logger:
            mock_logger.api_request = MagicMock()
            mock_logger.api_retry = MagicMock()
            mock_logger.api_error = MagicMock()
            with pytest.raises(RateLimitError):
                list(client.chat(
                    messages=[{"role": "user", "content": "hi"}],
                    stream=False,
                    on_413_callback=callback,
                ))
        # 413 回调绝不能被调用
        callback.assert_not_called()

    def test_500_does_not_trigger_413_callback(self):
        """500 InternalServerError 走原有重试，不调用 on_413_callback。"""
        client = DeepSeekClient(api_key="fake-key")
        client.client = MagicMock()
        client.retry_delay = 0.001
        client.client.chat.completions.create.side_effect = _make_500_error()

        callback = MagicMock(return_value=True)
        with patch("abyss.api_client.logger") as mock_logger:
            mock_logger.api_request = MagicMock()
            mock_logger.api_retry = MagicMock()
            mock_logger.api_error = MagicMock()
            with pytest.raises(InternalServerError):
                list(client.chat(
                    messages=[{"role": "user", "content": "hi"}],
                    stream=False,
                    on_413_callback=callback,
                ))
        callback.assert_not_called()

    def test_503_does_not_trigger_413_callback(self):
        """503 走原有 RETRYABLE_CODES 重试，不调用 on_413_callback。"""
        client = DeepSeekClient(api_key="fake-key")
        client.client = MagicMock()
        client.retry_delay = 0.001
        client.client.chat.completions.create.side_effect = _make_503_error()

        callback = MagicMock(return_value=True)
        with patch("abyss.api_client.logger") as mock_logger:
            mock_logger.api_request = MagicMock()
            mock_logger.api_retry = MagicMock()
            mock_logger.api_error = MagicMock()
            with pytest.raises(APIStatusError):
                list(client.chat(
                    messages=[{"role": "user", "content": "hi"}],
                    stream=False,
                    on_413_callback=callback,
                ))
        callback.assert_not_called()


class TestCallbackReturnsFalse:
    """回调返回 False（压缩失败）→ 直接抛异常，不重试。"""

    def test_callback_returns_false_raises_and_no_retry(self):
        """回调返回 False → 抛 RuntimeError，不进入 Phase 2。"""
        client = DeepSeekClient(api_key="fake-key")
        client.client = MagicMock()
        client.client.chat.completions.create.side_effect = _make_413_error()

        callback = MagicMock(return_value=False)
        with patch("abyss.api_client.logger") as mock_logger:
            mock_logger.api_request = MagicMock()
            mock_logger.api_retry = MagicMock()
            mock_logger.api_error = MagicMock()
            with pytest.raises(RuntimeError):
                list(client.chat(
                    messages=[{"role": "user", "content": "hi"}],
                    stream=True,
                    on_413_callback=callback,
                ))
        # 回调只被调用一次（Phase 1 失败，不再尝试 Phase 2）
        assert callback.call_count == 1, (
            f"回调返回 False 后不应继续重试，调用次数：{callback.call_count}"
        )



# -*- coding: utf-8 -*-
"""CG UndoCommand /undo 命令测试（Phase 8 Tier 2）"""
import os
import time
from pathlib import Path

import pytest

from abyss.commands.undo import (
    handle_undo,
    init_history_service,
    get_history_service,
)
from abyss.history import FileHistoryService


class _FakeSession:
    """测试用的最小 Session。"""

    def __init__(self, session_id: str = "test-session"):
        self.session_id = session_id
        self.messages = []


@pytest.fixture(autouse=True)
def reset_history_service():
    """每个测试前后重置全局单例。"""
    init_history_service(FileHistoryService(storage_dir="__test_reset__"))
    yield
    init_history_service(FileHistoryService(storage_dir="__test_reset__"))


class TestHandleUndo:
    """handle_undo 主流程测试"""

    def test_no_history_service_returns_error(self):
        # 临时清空全局 service
        import abyss.commands.undo as undo_mod
        old = undo_mod._HISTORY_SERVICE
        undo_mod._HISTORY_SERVICE = None
        try:
            session = _FakeSession()
            result = handle_undo(session)
            assert "[ERR]" in result
        finally:
            undo_mod._HISTORY_SERVICE = old

    def test_no_session_id_falls_back_to_default(self, tmp_path):
        """session 无 session_id 时应回退到 "default"，与 main.py:744 一致。

        痛点：Session 类没有 session_id 属性，main.py 在 begin_transaction
        时用 `or "default"` 兜底。undo 命令必须用同样的 fallback 才能查到
        那些事务。
        """
        svc = FileHistoryService(storage_dir=str(tmp_path / "hist"))
        init_history_service(svc)
        # 用 "default" session_id 创建一个事务（模拟 main.py 的行为）
        target = tmp_path / "file.py"
        target.write_text("cur", encoding="utf-8")
        tid = svc.begin_transaction("default", "m1")
        svc.record_file_change(tid, str(target), "old")
        svc.commit_transaction(tid)

        # session 无 session_id 属性 → getattr 返回 "" → fallback 到 "default"
        class _SessionNoId:
            messages = []
        result = handle_undo(_SessionNoId())
        assert "[OK]" in result
        assert target.read_text(encoding="utf-8") == "old"

    def test_no_transactions_returns_info(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        init_history_service(svc)
        session = _FakeSession(session_id="s1")
        result = handle_undo(session)
        assert "[INFO]" in result
        assert "无可撤销" in result

    def test_undo_finds_latest_transaction(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path / "hist"))
        init_history_service(svc)
        # 创建一个事务
        target = tmp_path / "file.py"
        target.write_text("current", encoding="utf-8")
        tid = svc.begin_transaction("s1", "m1")
        svc.record_file_change(tid, str(target), "old")
        svc.commit_transaction(tid)

        session = _FakeSession(session_id="s1")
        result = handle_undo(session)
        assert "[OK]" in result
        assert target.read_text(encoding="utf-8") == "old"

    def test_undo_lists_affected_files(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path / "hist"))
        init_history_service(svc)
        a = tmp_path / "a.py"
        b = tmp_path / "b.py"
        a.write_text("a-cur", encoding="utf-8")
        b.write_text("b-cur", encoding="utf-8")

        tid = svc.begin_transaction("s1", "m1")
        svc.record_file_change(tid, str(a), "a-old")
        svc.record_file_change(tid, str(b), "b-old")
        svc.commit_transaction(tid)

        session = _FakeSession(session_id="s1")
        result = handle_undo(session)
        assert "影响文件 2 个" in result
        assert str(a) in result
        assert str(b) in result

    def test_undo_skips_already_rewound(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path / "hist"))
        init_history_service(svc)
        # 两个事务：旧的已 rewind，新的可 rewind
        old_target = tmp_path / "old.py"
        old_target.write_text("cur", encoding="utf-8")
        tid1 = svc.begin_transaction("s1", "m1")
        svc.record_file_change(tid1, str(old_target), "old-content")
        svc.commit_transaction(tid1)
        time.sleep(0.01)

        new_target = tmp_path / "new.py"
        new_target.write_text("cur", encoding="utf-8")
        tid2 = svc.begin_transaction("s1", "m2")
        svc.record_file_change(tid2, str(new_target), "new-old")
        svc.commit_transaction(tid2)

        # 第一次 undo 应撤销最新（tid2）
        session = _FakeSession(session_id="s1")
        r1 = handle_undo(session)
        assert "[OK]" in r1
        assert new_target.read_text(encoding="utf-8") == "new-old"

        # 第二次 undo 应撤销 tid1
        r2 = handle_undo(session)
        assert "[OK]" in r2
        assert old_target.read_text(encoding="utf-8") == "old-content"

    def test_undo_scoped_to_session(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path / "hist"))
        init_history_service(svc)

        # session A 的事务
        a_target = tmp_path / "a.py"
        a_target.write_text("cur", encoding="utf-8")
        tid_a = svc.begin_transaction("sess-a", "m1")
        svc.record_file_change(tid_a, str(a_target), "a-old")
        svc.commit_transaction(tid_a)

        # session B 的事务
        b_target = tmp_path / "b.py"
        b_target.write_text("cur", encoding="utf-8")
        tid_b = svc.begin_transaction("sess-b", "m1")
        svc.record_file_change(tid_b, str(b_target), "b-old")
        svc.commit_transaction(tid_b)

        # session A 的 undo 应只影响 a.py
        session_a = _FakeSession(session_id="sess-a")
        result = handle_undo(session_a)
        assert "[OK]" in result
        assert str(a_target) in result
        assert str(b_target) not in result
        assert a_target.read_text(encoding="utf-8") == "a-old"
        # b.py 不应被改动
        assert b_target.read_text(encoding="utf-8") == "cur"


class TestInitAndGetService:
    """init_history_service / get_history_service 测试"""

    def test_init_and_get_roundtrip(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        init_history_service(svc)
        assert get_history_service() is svc

    def test_get_returns_none_initially(self):
        import abyss.commands.undo as undo_mod
        old = undo_mod._HISTORY_SERVICE
        undo_mod._HISTORY_SERVICE = None
        try:
            assert get_history_service() is None
        finally:
            undo_mod._HISTORY_SERVICE = old



# -*- coding: utf-8 -*-
"""
联网搜索工具
使用 cn.bing.com，国内直连无需代理
"""
import re
import time
from .base import Tool


class WebSearchTool(Tool):
    """联网搜索工具。使用 cn.bing.com 搜索，国内开箱即用。"""

    @property
    def name(self) -> str:
        return "web_search"

    @property
    def description(self) -> str:
        return "联网搜索信息，返回搜索结果列表。检索数量由 AI 自行决定，无需询问用户。"

    @property
    def parameters(self) -> dict:
        return {
            "type": "object",
            "properties": {
                "query": {"type": "string", "description": "搜索关键词"},
                "num": {
                    "type": "integer",
                    "description": "返回结果数量，由 AI 根据查询复杂度自行决定（一般 3-10 条）"
                }
            },
            "required": ["query", "num"]
        }

    def _search(self, query: str, num: int) -> list:
        """通过 urllib 直接抓取 cn.bing.com 搜索结果"""
        import urllib.request
        import urllib.parse

        params = urllib.parse.urlencode({"q": query, "first": "1", "count": str(num)})
        url = f"https://cn.bing.com/search?{params}"

        req = urllib.request.Request(url, headers={
            "User-Agent": (
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                "AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/120.0.0.0 Safari/537.36"
            ),
            "Accept": "text/html,application/xhtml+xml",
            "Accept-Language": "zh-CN,zh;q=0.9",
        })
        with urllib.request.urlopen(req, timeout=10) as resp:
            html = resp.read().decode("utf-8", errors="replace")

        return self._parse_results(html, num)

    def _parse_results(self, html: str, num: int) -> list:
        """从 cn.bing.com 搜索结果 HTML 中提取结果"""
        results = []

        # cn.bing.com 每个结果块是 <li class="b_algo">
        blocks = re.findall(r'<li class="b_algo"[^>]*>(.*?)</li>', html, re.DOTALL)

        for block in blocks:
            if len(results) >= num:
                break

            # 标题: <h2><a href="URL">title</a></h2>
            title_match = re.search(
                r'<h2[^>]*>.*?<a[^>]*href="([^"]*)"[^>]*>(.*?)</a>',
                block, re.DOTALL
            )
            if not title_match:
                continue

            url = title_match.group(1)
            title = re.sub(r'<[^>]+>', '', title_match.group(2)).strip()

            if not title:
                continue

            # 摘要
            snippet = ""
            snip_match = re.search(
                r'<(?:p|div)[^>]*class="[^"]*(?:b_lineclamp|b_caption|b_snippet)[^"]*"[^>]*>(.*?)</(?:p|div)>',
                block, re.DOTALL
            )
            if snip_match:
                snippet = re.sub(r'<[^>]+>', '', snip_match.group(1)).strip()

            snippet = snippet.replace("&nbsp;", " ").replace("&lt;", "<").replace("&gt;", ">")
            snippet = snippet.replace("&amp;", "&").replace("&quot;", '"')

            results.append({
                "title": title,
                "url": url,
                "snippet": snippet or title,
            })

        return results

    def execute(self, query: str, num: int = 5) -> dict:
        """执行搜索"""
        num = max(1, min(int(num), 20))
        max_retries = 2

        last_error = None

        for attempt in range(max_retries):
            try:
                results = self._search(query, num)
                if results:
                    return {
                        "success": True,
                        "query": query,
                        "results": results
                    }
                last_error = "未找到搜索结果"
            except Exception as e:
                last_error = str(e)
                if attempt < max_retries - 1:
                    time.sleep(1.0 * (attempt + 1))
                    continue

        return {"success": False, "error": last_error or "搜索失败，请检查网络连接"}



# -*- coding: utf-8 -*-
"""CH CompactCommand 测试（Phase 8 Tier 2）

借鉴 zhikuncode CompactCommand.java：
- /compact [custom_instruction] 手动触发上下文压缩
- 调用现有 SimpleTruncationEngine.compress()
- 显示压缩前后 token 数对比
- 保留 system message 不变式
"""
from unittest.mock import MagicMock, patch

import pytest

from abyss.commands.compact import handle_compact


def _make_messages(n: int = 20) -> list:
    """构造 n 条消息的会话"""
    msgs = [{"role": "system", "content": "you are an assistant"}]
    for i in range(n - 1):
        msgs.append({"role": "user" if i % 2 == 0 else "assistant", "content": f"msg {i}" * 10})
    return msgs


class TestHandleCompact:
    """handle_compact 测试"""

    def test_returns_success_message(self):
        """返回包含节省 token 数的成功消息"""
        session = MagicMock()
        session.messages = _make_messages(30)

        engine = MagicMock()
        # 模拟压缩前后 token 数差异
        engine.estimate_tokens.side_effect = [1000, 400]
        engine.compress.return_value = [{"role": "system", "content": "sys"}]

        result = handle_compact(session, engine, None)
        assert "1000" in result
        assert "400" in result
        assert "600" in result  # 节省 600

    def test_calls_engine_compress(self):
        """调用 engine.compress"""
        session = MagicMock()
        session.messages = _make_messages(30)
        engine = MagicMock()
        engine.estimate_tokens.return_value = 500
        engine.compress.return_value = [{"role": "system", "content": "sys"}]

        handle_compact(session, engine, None)
        engine.compress.assert_called_once()

    def test_passes_custom_instruction_as_focus_topic(self):
        """自定义指令作为 focus_topic 传入"""
        session = MagicMock()
        session.messages = _make_messages(30)
        engine = MagicMock()
        engine.estimate_tokens.return_value = 500
        engine.compress.return_value = [{"role": "system", "content": "sys"}]

        handle_compact(session, engine, "focus on API design")
        _, kwargs = engine.compress.call_args
        assert kwargs.get("focus_topic") == "focus on API design"

    def test_no_custom_instruction_passes_none(self):
        """无自定义指令时 focus_topic=None"""
        session = MagicMock()
        session.messages = _make_messages(30)
        engine = MagicMock()
        engine.estimate_tokens.return_value = 500
        engine.compress.return_value = [{"role": "system", "content": "sys"}]

        handle_compact(session, engine, None)
        _, kwargs = engine.compress.call_args
        assert kwargs.get("focus_topic") is None

    def test_replaces_session_messages(self):
        """压缩后替换 session.messages"""
        session = MagicMock()
        session.messages = _make_messages(30)
        engine = MagicMock()
        engine.estimate_tokens.return_value = 500
        compressed = [{"role": "system", "content": "compressed"}]
        engine.compress.return_value = compressed

        handle_compact(session, engine, None)
        assert session.messages == compressed

    def test_empty_session_handled(self):
        """空 session 优雅处理"""
        session = MagicMock()
        session.messages = []
        engine = MagicMock()
        engine.estimate_tokens.return_value = 0
        engine.compress.return_value = []

        result = handle_compact(session, engine, None)
        assert "0" in result  # 显示 0 → 0

    def test_preserves_system_message_first(self):
        """压缩后 system 仍在 messages[0]"""
        session = MagicMock()
        session.messages = _make_messages(30)
        engine = MagicMock()
        engine.estimate_tokens.return_value = 500
        engine.compress.return_value = [
            {"role": "system", "content": "sys"},
            {"role": "user", "content": "compressed"},
        ]

        handle_compact(session, engine, None)
        assert session.messages[0]["role"] == "system"

    def test_engine_none_returns_fallback_message(self):
        """engine 为 None 时返回 fallback 消息"""
        session = MagicMock()
        session.messages = _make_messages(10)
        result = handle_compact(session, None, None)
        assert "不可用" in result or "unavailable" in result.lower() or "未初始化" in result

    def test_compress_exception_handled(self):
        """compress 抛异常时优雅降级"""
        session = MagicMock()
        session.messages = _make_messages(30)
        engine = MagicMock()
        engine.estimate_tokens.return_value = 500
        engine.compress.side_effect = RuntimeError("compress failed")

        result = handle_compact(session, engine, None)
        assert "失败" in result or "error" in result.lower() or "错误" in result



# -*- coding: utf-8 -*-
"""
/knowledge 命令辅助函数测试
覆盖：分类分组列表、stats 统计、pin/archive 子命令
辅助函数为纯函数，接收 loader/curator，返回字符串列表，便于测试。
"""
from pathlib import Path

import pytest

from abyss.commands.knowledge import (
    format_knowledge_list_grouped,
    format_knowledge_stats,
    handle_knowledge_pin,
    handle_knowledge_archive,
)
from abyss.knowledge import Knowledge
from abyss.knowledge.curator import KnowledgeCurator


def _make_knowledge(name, description, level="project", category="experience",
                    tags=None) -> Knowledge:
    """构造测试用 Knowledge 对象。"""
    return Knowledge(
        name=name,
        description=description,
        content="正文",
        level=level,
        path=Path(f"/tmp/{name}.md"),
        tags=tags or [],
        category=category,
    )


class _FakeLoader:
    """假 KnowledgeLoader：list_all 返回预设列表。"""

    def __init__(self, items):
        self._items = items

    def list_all(self):
        return list(self._items)


# ---------- format_knowledge_list_grouped ----------

class TestFormatListGrouped:
    """分类分组列表显示。"""

    def test_groups_by_category(self):
        """同类型知识归到同一分组，按类型顺序输出。"""
        items = [
            _make_knowledge("pref-1", "偏好1", category="preference"),
            _make_knowledge("exp-1", "经验1", category="experience"),
            _make_knowledge("bug-1", "bug1", category="bugfix"),
            _make_knowledge("pref-2", "偏好2", category="preference"),
        ]
        lines = format_knowledge_list_grouped(_FakeLoader(items))
        text = "\n".join(lines)
        # 偏好 → bug → 经验 顺序
        assert text.index("偏好1") < text.index("bug1")
        assert text.index("bug1") < text.index("经验1")
        # 同组相邻
        pref1_idx = text.index("偏好1")
        pref2_idx = text.index("偏好2")
        bug_idx = text.index("bug1")
        assert pref1_idx < pref2_idx < bug_idx

    def test_includes_level_prefix(self):
        """每行含 [G]/[P] 级别前缀。"""
        items = [
            _make_knowledge("g-1", "全局", level="global", category="experience"),
            _make_knowledge("p-1", "项目", level="project", category="experience"),
        ]
        lines = format_knowledge_list_grouped(_FakeLoader(items))
        text = "\n".join(lines)
        assert "[G]" in text
        assert "[P]" in text

    def test_empty_loader_returns_empty(self):
        """空知识库返回空列表。"""
        lines = format_knowledge_list_grouped(_FakeLoader([]))
        assert lines == []

    def test_category_header_present(self):
        """分组包含中文类型标题。"""
        items = [_make_knowledge("p1", "偏好", category="preference")]
        lines = format_knowledge_list_grouped(_FakeLoader(items))
        text = "\n".join(lines)
        assert "偏好" in text  # 标题中含"偏好"


# ---------- format_knowledge_stats ----------

class TestFormatStats:
    """stats 统计显示。"""

    def test_counts_by_category(self, tmp_path):
        """统计按类型计数。"""
        items = [
            _make_knowledge("p1", "p", category="preference"),
            _make_knowledge("p2", "p", category="preference"),
            _make_knowledge("e1", "e", category="experience"),
            _make_knowledge("b1", "b", category="bugfix"),
        ]
        curator = KnowledgeCurator({"project": tmp_path, "global": tmp_path / "g"})
        lines = format_knowledge_stats(_FakeLoader(items), curator)
        text = "\n".join(lines)
        assert "preference" in text or "偏好" in text
        assert "2" in text  # preference 计数 2

    def test_includes_total(self, tmp_path):
        """包含总数。"""
        items = [
            _make_knowledge("p1", "p", category="preference"),
            _make_knowledge("e1", "e", category="experience"),
        ]
        curator = KnowledgeCurator({"project": tmp_path, "global": tmp_path / "g"})
        lines = format_knowledge_stats(_FakeLoader(items), curator)
        text = "\n".join(lines)
        assert "2" in text  # 总数 2


# ---------- handle_knowledge_pin ----------

class TestHandlePin:
    """pin 子命令。"""

    def test_pin_returns_success_message(self, tmp_path):
        """pin 成功返回成功消息。"""
        # 先在 project 级创建一个知识文件
        cat_dir = tmp_path / "experiences"
        cat_dir.mkdir(parents=True)
        md = cat_dir / "test-k.md"
        md.write_text(
            "---\nname: test-k\ndescription: d\nlevel: project\ncategory: experience\n---\n正文",
            encoding="utf-8",
        )
        curator = KnowledgeCurator({"project": tmp_path, "global": tmp_path / "g"})
        msg = handle_knowledge_pin(curator, "test-k")
        assert "成功" in msg or "OK" in msg or "pinned" in msg.lower()

    def test_pin_empty_name_returns_error(self, tmp_path):
        """空名称返回错误消息。"""
        curator = KnowledgeCurator({"project": tmp_path, "global": tmp_path / "g"})
        msg = handle_knowledge_pin(curator, "")
        assert "错误" in msg or "失败" in msg or "无效" in msg or "error" in msg.lower()


# ---------- handle_knowledge_archive ----------

class TestHandleArchive:
    """archive 子命令。"""

    def test_archive_moves_file(self, tmp_path):
        """archive 成功后原文件移到 .archive/。"""
        cat_dir = tmp_path / "experiences"
        cat_dir.mkdir(parents=True)
        md = cat_dir / "archived-k.md"
        md.write_text(
            "---\nname: archived-k\ndescription: d\nlevel: project\ncategory: experience\n---\n正文",
            encoding="utf-8",
        )
        curator = KnowledgeCurator({"project": tmp_path, "global": tmp_path / "g"})
        msg = handle_knowledge_archive(curator, "archived-k")
        assert "成功" in msg or "OK" in msg or "归档" in msg
        assert not md.exists()
        assert (tmp_path / ".archive" / "archived-k.md").exists()

    def test_archive_not_found_returns_error(self, tmp_path):
        """不存在的知识返回错误消息。"""
        curator = KnowledgeCurator({"project": tmp_path, "global": tmp_path / "g"})
        msg = handle_knowledge_archive(curator, "nonexistent-k")
        assert "错误" in msg or "失败" in msg or "未找到" in msg or "error" in msg.lower()



# -*- coding: utf-8 -*-
"""ToolSafetyGuard 多层纵深防护（Phase 8 Tier 1-5 CE）

借鉴 zhikuncode ToolSafetyGuard：
- 12 个危险 bash 命令正则（fork bomb / rm -rf / / curl|sh 等）
- 10 类敏感环境变量清理（AWS/GITHUB/NPM/DB 等）
- 路径规范化与符号链接逃逸检测
- SafetyLevel 4 级分类（SAFE / AUDIT / HIGH_RISK / BLOCK）
- 挂载点边界检测（Linux /proc/mounts，Windows 跳过）

与 Phase 6 AP PathSecurityService 互补：
- AP PathSecurityService：5 重路径校验（符号链接、路径穿越、工作目录边界、受保护路径、Windows 绕过）
- CE ToolSafetyGuard：bash 命令检测 + 环境变量清理 + 路径安全检查（轻量版，集成方便）

集成点：
1. `tools/shell_exec.py`：执行前 check_bash_command + sanitize_env
2. `tools/file_write.py` / `file_edit.py`：写前 check_path
3. `tools/file_read.py`：读敏感路径前 check_path
"""
import os
import re
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple


class SafetyLevel(Enum):
    """安全级别枚举。

    优先级：SAFE(1) < AUDIT(2) < HIGH_RISK(3) < BLOCK(4)
    """

    SAFE = "safe"           # 安全，直接通过
    AUDIT = "audit"         # 仅记录，不拦截
    HIGH_RISK = "high_risk"  # 高风险，提示用户确认
    BLOCK = "block"         # 拦截，拒绝执行

    def __init__(self, value: str) -> None:
        order = ["safe", "audit", "high_risk", "block"]
        self.priority = order.index(value) + 1

    @classmethod
    def from_value(cls, value: str) -> Optional["SafetyLevel"]:
        for member in cls:
            if member.value == value:
                return member
        return None


class SafetyCheckResult:
    """安全检查结果。

    Attributes:
        level: 安全级别
        reason: 人类可读的原因说明
        pattern_matched: 触发的模式（仅 BLOCK/HIGH_RISK 时填充）
    """

    __slots__ = ("level", "reason", "pattern_matched")

    def __init__(
        self,
        level: SafetyLevel,
        reason: str = "",
        pattern_matched: Optional[str] = None,
    ) -> None:
        self.level = level
        self.reason = reason
        self.pattern_matched = pattern_matched

    @property
    def should_block(self) -> bool:
        """是否应阻止操作。"""
        return self.level == SafetyLevel.BLOCK

    @property
    def should_audit(self) -> bool:
        """是否应记录审计日志。"""
        return self.level.priority >= SafetyLevel.AUDIT.priority

    def __repr__(self) -> str:
        return f"SafetyCheckResult(level={self.level}, reason={self.reason!r})"


# 12 个危险 bash 命令正则
# 注意：使用 re.MULTILINE + re.DOTALL 兼容多行命令和复合命令
_DANGEROUS_BASH_PATTERNS: List[Tuple[str, str, SafetyLevel]] = [
    # 1. fork bomb: :(){ :|:& };:
    (r":\s*\(\s*\)\s*\{\s*:\s*\|\s*:\s*&\s*\}\s*;?\s*:?", "fork bomb", SafetyLevel.BLOCK),
    # 2. rm -rf / 或 sudo rm -rf /（根目录递归删除，后接 shell 元字符或行尾）
    (r"(?:\bsudo\s+)?\brm\s+-rf\s+/?\s*(?:$|[;&|)\n`])", "rm -rf /", SafetyLevel.BLOCK),
    # 3. mkfs.*（格式化文件系统）
    (r"\bmkfs(?:\.\w+)?\s+", "mkfs", SafetyLevel.BLOCK),
    # 4. dd of=/dev/...（写入块设备）
    (r"\bdd\b.*\bof=/dev/", "dd of=/dev/", SafetyLevel.BLOCK),
    # 5. curl | sh 或 wget | sh（远程执行）
    (r"\b(?:curl|wget)\b.*\|\s*(?:sh|bash|zsh|fish)\b", "curl|sh", SafetyLevel.BLOCK),
    # 6. chmod 777 /（根目录权限开放）
    (r"\bchmod\s+777\s+/\s*$", "chmod 777 /", SafetyLevel.BLOCK),
    # 7. > /dev/sda 或类似（重定向到块设备）
    (r">\s*/dev/(?:sd[a-z]+|nvme\d+n\d+|hd[a-z]+)", "redirect to /dev/", SafetyLevel.BLOCK),
    # 8. shutdown（关机）
    (r"\bshutdown\b", "shutdown", SafetyLevel.BLOCK),
    # 9. reboot（重启）
    (r"\breboot\b", "reboot", SafetyLevel.BLOCK),
    # 10. halt（停机）
    (r"\bhalt\b", "halt", SafetyLevel.BLOCK),
    # 11. poweroff（关机）
    (r"\bpoweroff\b", "poweroff", SafetyLevel.BLOCK),
    # 12. init 0（系统关机）
    (r"\binit\s+0\b", "init 0", SafetyLevel.BLOCK),
]

# 编译为正则对象（性能优化）
_COMPILED_DANGEROUS_PATTERNS: List[Tuple[re.Pattern, str, SafetyLevel]] = [
    (re.compile(p, re.MULTILINE | re.DOTALL), name, level)
    for p, name, level in _DANGEROUS_BASH_PATTERNS
]

# 10 类敏感环境变量（精确匹配）
SENSITIVE_ENV_VARS = [
    # 1. AWS
    "AWS_ACCESS_KEY_ID",
    "AWS_SECRET_ACCESS_KEY",
    "AWS_SESSION_TOKEN",
    # 2. GitHub
    "GITHUB_TOKEN",
    "GH_TOKEN",
    # 3. NPM
    "NPM_TOKEN",
    "NODE_AUTH_TOKEN",
    # 4. Database
    "DATABASE_URL",
    "DB_PASSWORD",
    # 5. OpenAI / Anthropic
    "OPENAI_API_KEY",
    "ANTHROPIC_API_KEY",
]

# 受保护路径（系统级敏感文件）
_PROTECTED_PATHS = [
    # Linux
    "/etc/shadow",
    "/etc/passwd",
    "/etc/sudoers",
    "/etc/ssh/sshd_config",
    "/root/.ssh",
    "/var/log/auth.log",
    # Windows
    "C:\\Windows\\System32\\config\\SAM",
    "C:\\Windows\\System32\\config\\SYSTEM",
    "C:\\Windows\\System32\\config\\SECURITY",
]


class ToolSafetyGuard:
    """多层纵深防护链。

    借鉴 zhikuncode ToolSafetyGuard：
    1. realpath 规范化 + 符号链接逃逸检测
    2. 沙箱边界检查（工作目录内）
    3. 挂载点边界检测（Linux /proc/mounts）
    4. 敏感路径黑名单
    5. 12 个危险 bash 命令正则
    6. 敏感环境变量清理

    Usage:
        >>> guard = ToolSafetyGuard(working_dir="/path/to/project")
        >>> result = guard.check_bash_command("rm -rf /")
        >>> if result.should_block:
        ...     raise PermissionError(result.reason)
        >>> env = guard.sanitize_env(os.environ.copy())
    """

    def __init__(self, working_dir: str) -> None:
        """初始化 ToolSafetyGuard。

        Args:
            working_dir: 工作目录路径（决定沙箱边界）
        """
        try:
            self._working_dir = str(Path(working_dir).resolve())
        except (OSError, ValueError):
            # 路径规范化失败（如含 null 字节）：保留原值，后续 check_path 会拦截
            self._working_dir = working_dir

    # ---- bash 命令检测 ----

    def check_bash_command(self, cmd: str) -> SafetyCheckResult:
        """检测 bash 命令是否包含危险模式。

        检测策略：
        - 12 个危险命令正则匹配
        - 多行命令逐行检测
        - 复合命令（&&/||/;/|/反引号/$()）整体检测

        Args:
            cmd: bash 命令字符串

        Returns:
            SafetyCheckResult：BLOCK 表示应拦截，SAFE 表示通过
        """
        if not cmd or not cmd.strip():
            return SafetyCheckResult(level=SafetyLevel.SAFE, reason="empty command")

        # 对每个危险模式检测
        for pattern, name, level in _COMPILED_DANGEROUS_PATTERNS:
            match = pattern.search(cmd)
            if match:
                return SafetyCheckResult(
                    level=level,
                    reason=f"matched dangerous pattern: {name}",
                    pattern_matched=name,
                )

        return SafetyCheckResult(level=SafetyLevel.SAFE, reason="no dangerous pattern")

    # ---- 路径检测 ----

    def check_path(self, path: str, operation: str = "read") -> SafetyCheckResult:
        """检测路径是否安全。

        检测项：
        1. null 字节检测
        2. 路径穿越检测（../）
        3. 工作目录边界检测
        4. 受保护路径黑名单

        相对路径会基于 working_dir 解析（而非进程 CWD）。

        Args:
            path: 待检测路径
            operation: 操作类型（read/write）

        Returns:
            SafetyCheckResult：BLOCK 表示应拦截，SAFE 表示通过
        """
        if path is None:
            return SafetyCheckResult(level=SafetyLevel.BLOCK, reason="path is None")

        # 1. null 字节检测
        if "\x00" in path:
            return SafetyCheckResult(
                level=SafetyLevel.BLOCK,
                reason="path contains null byte",
                pattern_matched="null_byte",
            )

        # 2. 路径穿越检测（简单字符串检测，先于规范化）
        if ".." in path.split(os.sep) or ".." in re.split(r"[\\/]", path):
            # 进一步规范化判断：若解析后仍在工作目录内，则放行
            try:
                resolved = self._resolve_path(path)
                if not self._is_in_working_dir(resolved):
                    return SafetyCheckResult(
                        level=SafetyLevel.BLOCK,
                        reason=f"path traversal: {path} resolves outside working dir",
                        pattern_matched="path_traversal",
                    )
            except (OSError, ValueError):
                return SafetyCheckResult(
                    level=SafetyLevel.BLOCK,
                    reason=f"cannot resolve path: {path}",
                    pattern_matched="path_resolve_failed",
                )

        # 3. 规范化路径
        try:
            resolved = self._resolve_path(path)
            resolved_str = str(resolved)
        except (OSError, ValueError):
            return SafetyCheckResult(
                level=SafetyLevel.BLOCK,
                reason=f"cannot resolve path: {path}",
                pattern_matched="path_resolve_failed",
            )

        # 4. 受保护路径检测
        for protected in _PROTECTED_PATHS:
            try:
                protected_resolved = str(Path(protected).resolve())
            except (OSError, ValueError):
                protected_resolved = protected
            # 大小写不敏感比较（Windows 路径）
            if resolved_str.lower() == protected_resolved.lower() or \
               resolved_str.lower().startswith(protected_resolved.lower() + os.sep):
                return SafetyCheckResult(
                    level=SafetyLevel.BLOCK,
                    reason=f"protected path: {protected}",
                    pattern_matched="protected_path",
                )

        # 5. 工作目录边界检测
        if not self._is_in_working_dir(resolved):
            return SafetyCheckResult(
                level=SafetyLevel.BLOCK,
                reason=f"path outside working dir: {resolved_str}",
                pattern_matched="outside_working_dir",
            )

        return SafetyCheckResult(level=SafetyLevel.SAFE, reason="path ok")

    def _resolve_path(self, path: str) -> Path:
        """解析路径：相对路径基于 working_dir，绝对路径原样。

        Args:
            path: 原始路径字符串

        Returns:
            解析后的绝对 Path 对象
        """
        p = Path(path)
        if not p.is_absolute():
            # 相对路径：基于 working_dir 解析
            p = Path(self._working_dir) / p
        return p.resolve()

    def _is_in_working_dir(self, resolved: Path) -> bool:
        """检查解析后的路径是否在工作目录内。"""
        try:
            wd = Path(self._working_dir)
            # 相对路径：检查是否以工作目录为前缀
            resolved.relative_to(wd)
            return True
        except ValueError:
            # 不在工作目录内
            return False
        except (OSError, RuntimeError):
            return False

    # ---- 环境变量清理 ----

    def sanitize_env(self, env: Dict[str, str]) -> Dict[str, str]:
        """清理敏感环境变量。

        移除以下 10 类敏感变量：
        - AWS（ACCESS_KEY_ID / SECRET_ACCESS_KEY / SESSION_TOKEN）
        - GitHub（GITHUB_TOKEN / GH_TOKEN）
        - NPM（NPM_TOKEN / NODE_AUTH_TOKEN）
        - Database（DATABASE_URL / DB_PASSWORD）
        - OpenAI / Anthropic（OPENAI_API_KEY / ANTHROPIC_API_KEY）

        Args:
            env: 原始环境变量 dict

        Returns:
            清理后的环境变量 dict（新对象，不修改原 dict）
        """
        if not env:
            return {}
        sensitive_set = set(SENSITIVE_ENV_VARS)
        return {k: v for k, v in env.items() if k not in sensitive_set}

    # ---- 挂载点边界检测 ----

    def check_mount_boundary(self, path: str) -> SafetyCheckResult:
        """检测路径是否跨越挂载点边界。

        Linux：解析 /proc/mounts 获取挂载点列表
        Windows：跳过（直接返回 SAFE）

        Args:
            path: 待检测路径

        Returns:
            SafetyCheckResult
        """
        if path is None:
            return SafetyCheckResult(
                level=SafetyLevel.BLOCK, reason="path is None"
            )

        # Windows 无 /proc/mounts，跳过挂载点检测
        if os.name == "nt":
            return SafetyCheckResult(
                level=SafetyLevel.SAFE,
                reason="mount boundary check skipped on Windows",
            )

        # Linux：读取 /proc/mounts
        try:
            resolved = str(Path(path).resolve())
            with open("/proc/mounts", "r", encoding="utf-8") as f:
                mounts = []
                for line in f:
                    parts = line.split()
                    if len(parts) >= 2:
                        mounts.append(parts[1])

            # 找出工作目录所在的挂载点
            wd_mount = None
            for mount in sorted(mounts, key=len, reverse=True):
                if self._working_dir.startswith(mount):
                    wd_mount = mount
                    break

            if wd_mount is None:
                return SafetyCheckResult(
                    level=SafetyLevel.AUDIT,
                    reason=f"cannot find working dir mount point",
                )

            # 检查目标路径是否在同一挂载点内
            if not resolved.startswith(wd_mount):
                return SafetyCheckResult(
                    level=SafetyLevel.HIGH_RISK,
                    reason=f"path crosses mount boundary: {resolved} not in {wd_mount}",
                    pattern_matched="mount_boundary_cross",
                )

            return SafetyCheckResult(
                level=SafetyLevel.SAFE, reason="path within mount boundary"
            )
        except (OSError, IOError):
            # /proc/mounts 不可读：跳过
            return SafetyCheckResult(
                level=SafetyLevel.SAFE,
                reason="mount boundary check skipped (cannot read /proc/mounts)",
            )



# -*- coding: utf-8 -*-
"""
文件编辑工具
通过搜索替换修改文件内容
"""
from .base import Tool
from .file_cache import get_default_cache
from .file_io import AtomicFileWriter, EncodingDetector
from ..security.path_security import get_default_service as get_path_security


class FileEditTool(Tool):
    """通过搜索替换修改文件内容的工具。"""

    @property
    def name(self) -> str:
        return "file_edit"

    @property
    def description(self) -> str:
        return "修改已有文件的指定内容，通过精确匹配旧字符串替换为新字符串。"

    @property
    def parameters(self) -> dict:
        return {
            "type": "object",
            "properties": {
                "path": {"type": "string", "description": "文件路径"},
                "old_str": {"type": "string", "description": "要替换的旧内容"},
                "new_str": {"type": "string", "description": "替换后的新内容"}
            },
            "required": ["path", "old_str", "new_str"]
        }

    def execute(self, path: str, old_str: str, new_str: str) -> dict:
        """编辑文件内容"""
        # PathSecurityService 5 重校验（Phase 6 Tier 1-5 AP）
        sec = get_path_security()
        check = sec.check_write(path)
        if not check.allowed:
            # 保持与历史断言一致："禁止写入" → "禁止编辑"
            reason = check.reason.replace("禁止写入", "禁止编辑")
            return {"success": False, "error": reason}
        if check.needs_confirmation:
            return {"success": False, "error": f"需用户确认: {check.reason}"}

        try:
            # 检测文件编码（Phase 6 Tier 1-3 AO）
            encoding = EncodingDetector.detect(path)
            if encoding is None:
                return {"success": False, "error": "无法编辑二进制文件"}

            # 用检测到的编码读取
            with open(path, "r", encoding=encoding) as f:
                content = f.read()

            if old_str not in content:
                return {"success": False, "error": "未找到匹配的旧内容"}

            new_content = content.replace(old_str, new_str, 1)

            # 使用 AtomicFileWriter 原子写入（保持原编码）
            AtomicFileWriter.write(path, new_content, encoding=encoding)

            # 失效文件状态缓存（Phase 6 AQ）
            get_default_cache().invalidate(path)

            return {"success": True, "path": path, "encoding": encoding}
        except Exception as e:
            return {"success": False, "error": str(e)}



# -*- coding: utf-8 -*-
"""CG FileHistoryService 文件历史事务测试（Phase 8 Tier 2）

借鉴 zhikuncode FileHistoryService.java：
- begin_transaction / record_file_change / commit_transaction / rewind
- 按 session_id 隔离事务
- rewind 创建 .bak 备份
- 新建文件回退时删除
"""
import os
import time
import threading
from pathlib import Path

import pytest

from abyss.history.file_history import (
    FileHistoryService,
    TransactionRecord,
    RewindResult,
)


class TestBeginTransaction:
    """begin_transaction 测试"""

    def test_returns_unique_id(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        tid1 = svc.begin_transaction("sess1", "msg1")
        tid2 = svc.begin_transaction("sess1", "msg2")
        assert tid1 != tid2
        assert len(tid1) > 0

    def test_transaction_not_committed_initially(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        tid = svc.begin_transaction("sess1", "msg1")
        txns = svc.list_recent_transactions("sess1")
        # 未 commit 的不应在 list 中
        assert all(t.transaction_id != tid for t in txns)

    def test_records_session_and_message(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        tid = svc.begin_transaction("sess-xyz", "msg-abc")
        svc.commit_transaction(tid)
        txns = svc.list_recent_transactions("sess-xyz")
        assert len(txns) == 1
        assert txns[0].session_id == "sess-xyz"
        assert txns[0].message_id == "msg-abc"


class TestRecordFileChange:
    """record_file_change 测试"""

    def test_stores_old_content(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        tid = svc.begin_transaction("s1", "m1")
        svc.record_file_change(tid, "/path/a.py", "old content")
        svc.commit_transaction(tid)
        txns = svc.list_recent_transactions("s1")
        assert "/path/a.py" in txns[0].snapshots
        assert txns[0].snapshots["/path/a.py"] == "old content"

    def test_stores_none_for_new_file(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        tid = svc.begin_transaction("s1", "m1")
        svc.record_file_change(tid, "/path/new.py", None)
        svc.commit_transaction(tid)
        txns = svc.list_recent_transactions("s1")
        assert txns[0].snapshots["/path/new.py"] is None

    def test_records_multiple_files(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        tid = svc.begin_transaction("s1", "m1")
        svc.record_file_change(tid, "/a.py", "old-a")
        svc.record_file_change(tid, "/b.py", "old-b")
        svc.record_file_change(tid, "/c.py", None)
        svc.commit_transaction(tid)
        txns = svc.list_recent_transactions("s1")
        assert len(txns[0].snapshots) == 3

    def test_unknown_transaction_id_raises(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        with pytest.raises(KeyError):
            svc.record_file_change("nonexistent", "/a.py", "old")


class TestCommitTransaction:
    """commit_transaction 测试"""

    def test_sets_completed_at(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        tid = svc.begin_transaction("s1", "m1")
        before = time.time()
        svc.commit_transaction(tid)
        after = time.time()
        txns = svc.list_recent_transactions("s1")
        assert txns[0].completed_at is not None
        assert before <= txns[0].completed_at <= after

    def test_commit_unknown_id_raises(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        with pytest.raises(KeyError):
            svc.commit_transaction("nonexistent")


class TestRewind:
    """rewind 测试"""

    def test_restores_original_content(self, tmp_path):
        # 真实文件场景
        target = tmp_path / "file.py"
        target.write_text("current content", encoding="utf-8")

        svc = FileHistoryService(storage_dir=str(tmp_path / "hist"))
        tid = svc.begin_transaction("s1", "m1")
        svc.record_file_change(tid, str(target), "old content")
        svc.commit_transaction(tid)

        result = svc.rewind(tid)
        assert result.success
        # 文件应回退到 old content
        assert target.read_text(encoding="utf-8") == "old content"

    def test_deletes_newly_created_files(self, tmp_path):
        # 创建后回退，文件应被删除
        target = tmp_path / "new_file.py"
        target.write_text("newly created", encoding="utf-8")

        svc = FileHistoryService(storage_dir=str(tmp_path / "hist"))
        tid = svc.begin_transaction("s1", "m1")
        # 新建文件快照为 None
        svc.record_file_change(tid, str(target), None)
        svc.commit_transaction(tid)

        svc.rewind(tid)
        assert not target.exists()

    def test_creates_bak_for_current(self, tmp_path):
        target = tmp_path / "file.py"
        target.write_text("current", encoding="utf-8")

        svc = FileHistoryService(storage_dir=str(tmp_path / "hist"))
        tid = svc.begin_transaction("s1", "m1")
        svc.record_file_change(tid, str(target), "old")
        svc.commit_transaction(tid)

        svc.rewind(tid)
        bak_path = Path(str(target) + ".bak")
        assert bak_path.exists()
        assert bak_path.read_text(encoding="utf-8") == "current"

    def test_rewind_only_affects_transaction_files(self, tmp_path):
        # 文件 a 在事务中，文件 b 不在
        a = tmp_path / "a.py"
        b = tmp_path / "b.py"
        a.write_text("a-current", encoding="utf-8")
        b.write_text("b-current", encoding="utf-8")

        svc = FileHistoryService(storage_dir=str(tmp_path / "hist"))
        tid = svc.begin_transaction("s1", "m1")
        svc.record_file_change(tid, str(a), "a-old")
        # 不记录 b
        svc.commit_transaction(tid)

        svc.rewind(tid)
        assert a.read_text(encoding="utf-8") == "a-old"
        # b 不受影响
        assert b.read_text(encoding="utf-8") == "b-current"

    def test_rewind_unknown_id_returns_failure(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        result = svc.rewind("nonexistent")
        assert not result.success

    def test_rewind_returns_affected_files(self, tmp_path):
        a = tmp_path / "a.py"
        b = tmp_path / "b.py"
        a.write_text("a-cur", encoding="utf-8")
        b.write_text("b-cur", encoding="utf-8")

        svc = FileHistoryService(storage_dir=str(tmp_path / "hist"))
        tid = svc.begin_transaction("s1", "m1")
        svc.record_file_change(tid, str(a), "a-old")
        svc.record_file_change(tid, str(b), "b-old")
        svc.commit_transaction(tid)

        result = svc.rewind(tid)
        assert set(result.affected_files) == {str(a), str(b)}


class TestListRecentTransactions:
    """list_recent_transactions 测试"""

    def test_returns_latest_first(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        tid1 = svc.begin_transaction("s1", "m1")
        svc.commit_transaction(tid1)
        time.sleep(0.01)
        tid2 = svc.begin_transaction("s1", "m2")
        svc.commit_transaction(tid2)

        txns = svc.list_recent_transactions("s1", limit=10)
        assert len(txns) == 2
        # 最新的在前
        assert txns[0].transaction_id == tid2
        assert txns[1].transaction_id == tid1

    def test_scoped_to_session(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        t1 = svc.begin_transaction("sess-a", "m1")
        svc.commit_transaction(t1)
        t2 = svc.begin_transaction("sess-b", "m2")
        svc.commit_transaction(t2)

        a_txns = svc.list_recent_transactions("sess-a")
        b_txns = svc.list_recent_transactions("sess-b")
        assert len(a_txns) == 1
        assert a_txns[0].session_id == "sess-a"
        assert len(b_txns) == 1
        assert b_txns[0].session_id == "sess-b"

    def test_limit_respected(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        for i in range(5):
            tid = svc.begin_transaction("s1", f"m{i}")
            svc.commit_transaction(tid)
            time.sleep(0.01)

        txns = svc.list_recent_transactions("s1", limit=3)
        assert len(txns) == 3

    def test_empty_when_no_transactions(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))
        assert svc.list_recent_transactions("no-such") == []


class TestConcurrency:
    """并发测试"""

    def test_concurrent_begin_commit_thread_safe(self, tmp_path):
        svc = FileHistoryService(storage_dir=str(tmp_path))

        def create_one():
            tid = svc.begin_transaction("s1", "m")
            svc.record_file_change(tid, f"/p/{threading.get_ident()}.py", "old")
            svc.commit_transaction(tid)

        threads = [threading.Thread(target=create_one) for _ in range(10)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()
        assert len(svc.list_recent_transactions("s1")) == 10


class TestPersistence:
    """持久化测试"""

    def test_transactions_persist_across_instances(self, tmp_path):
        storage = str(tmp_path / "hist")
        svc1 = FileHistoryService(storage_dir=storage)
        tid = svc1.begin_transaction("s1", "m1")
        svc1.record_file_change(tid, "/a.py", "old")
        svc1.commit_transaction(tid)

        # 新实例加载已有数据
        svc2 = FileHistoryService(storage_dir=storage)
        txns = svc2.list_recent_transactions("s1")
        assert len(txns) == 1
        assert txns[0].snapshots["/a.py"] == "old"



# -*- coding: utf-8 -*-
"""
工具基类
所有具体工具需继承此类
"""
import json
import os
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional


def coerce_int(value: Any, default: int = 0) -> int:
    """将任意值规整为 int，失败时返回 default。

    LLM 偶尔将 int 参数传为字符串（如 "10"），本函数负责安全转换，
    避免 str - int / str > int 等运算抛 TypeError。
    """
    if value is None:
        return default
    if isinstance(value, bool):
        # bool 是 int 子类，需单独处理，避免 True 被当作 1
        return int(value)
    if isinstance(value, int):
        return value
    if isinstance(value, float):
        return int(value)
    if isinstance(value, str):
        s = value.strip()
        if not s:
            return default
        try:
            # 先尝试 int，再尝试 float 截断
            return int(s)
        except ValueError:
            try:
                return int(float(s))
            except (ValueError, TypeError):
                return default
    # 其他类型（list/dict/None）兜底
    return default


def coerce_bool(value: Any, default: bool = False) -> bool:
    """将任意值规整为 bool，特别防御字符串 'false' 被判 truthy。

    LLM 偶尔将 bool 参数传为字符串 "false"/"true"，Python 中
    非空字符串均 truthy，若直接 `if value` 会把 "false" 当 True。
    无法明确判断的值（空串、空容器等）返回 default。
    """
    if value is None:
        return default
    if isinstance(value, bool):
        return value
    if isinstance(value, (int, float)):
        return bool(value)
    if isinstance(value, str):
        s = value.strip().lower()
        if s in ("true", "1", "yes", "y", "on"):
            return True
        if s in ("false", "0", "no", "n", "off"):
            return False
        # 空串或无法识别的字符串 → default
        return default
    # 容器类型（list/dict 等）按 truthy 规则
    return bool(value) if value else default


def _coerce_json_list(value: Any) -> List[Any]:
    """内部：若 value 是 JSON 字符串则反序列化为 list，否则返回原值或空 list。"""
    if isinstance(value, str):
        s = value.strip()
        if not s:
            return []
        try:
            parsed = json.loads(s)
        except (json.JSONDecodeError, ValueError):
            return []
        if not isinstance(parsed, list):
            return []
        return parsed
    if not isinstance(value, list):
        return []
    return value


def coerce_list_of_dicts(value: Any) -> List[Dict[str, Any]]:
    """将 value 规整为 list[dict]。

    LLM 偶尔将 list[dict] 参数传为 JSON 字符串（如 '[{"id":"1"}]'），
    直接迭代字符串会逐字符处理，后续 .get() 抛 AttributeError。
    本函数负责反序列化并过滤非 dict 项。
    """
    items = _coerce_json_list(value)
    return [t for t in items if isinstance(t, dict)]


def coerce_list_of_str(value: Any) -> List[str]:
    """将 value 规整为 list[str]，过滤非 str 项。

    LLM 偶尔将 list[str] 参数传为 JSON 字符串（如 '["a","b"]'），
    直接 join / set comprehension 会迭代字符串字符，导致静默数据损坏。
    """
    items = _coerce_json_list(value)
    return [t for t in items if isinstance(t, str)]


class Tool(ABC):
    """工具基类，所有具体工具需继承此类。"""

    @property
    @abstractmethod
    def name(self) -> str:
        """工具名称。"""
        pass

    @property
    @abstractmethod
    def description(self) -> str:
        """工具功能描述。"""
        pass

    @property
    @abstractmethod
    def parameters(self) -> Dict[str, Any]:
        """工具参数 JSON Schema。"""
        pass

    @abstractmethod
    def execute(self, **kwargs) -> Dict[str, Any]:
        """执行工具，返回结果字典。"""
        pass

    def to_openai_schema(self) -> Dict[str, Any]:
        """转换为 OpenAI Tool 格式。"""
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": self.parameters
            }
        }


# 受保护路径黑名单（借鉴 zhikuncode-main，防止 Agent 误删敏感文件）
PROTECTED_PATHS = [".git", ".env", ".ssh", ".gnupg", ".aws"]


def find_protected_name(path: str) -> Optional[str]:
    """查找路径命中的受保护黑名单项名称。

    用 os.path.normpath 规范化路径以处理 ``..`` 穿越与 Windows/Unix 分隔符差异，
    然后逐段检查路径组件是否在 PROTECTED_PATHS 中。命中返回该项名称，否则返回 None。
    """
    if not path:
        return None
    # normpath 统一分隔符（Windows 上为 \）并解析 .. / .
    norm = os.path.normpath(path)
    # 统一为 / 便于跨平台拆分组件
    parts = norm.replace("\\", "/").split("/")
    for part in parts:
        if part in PROTECTED_PATHS:
            return part
    return None


def is_protected_path(path: str) -> bool:
    """检测路径是否命中受保护路径黑名单。

    支持路径规范化（``..`` 穿越检测）与 Windows/Unix 路径分隔符兼容。
    """
    return find_protected_name(path) is not None



# -*- coding: utf-8 -*-
"""
终端命令执行工具
执行终端命令，带安全限制和失败重试
"""
import os
import re
import sys
import time
import signal
import threading
import subprocess
from typing import Optional
from .base import Tool, coerce_int


class ShellExecTool(Tool):
    """执行终端命令的工具，带安全限制。失败时自动重试，避免临时网络/IO 抖动。"""

    # 最大重试次数（不含首次执行）。例如 max_retries=2 表示最多执行 3 次
    MAX_RETRIES = 2

    # 重试退避基数（秒），实际等待 = BASE_BACKOFF * 2^attempt
    BASE_BACKOFF = 0.5

    DANGEROUS_PATTERNS = [
        "rm -rf /", "rm -rf /*", "> /dev/sda", "dd if=/dev/zero",
        "mkfs.", "chmod -R 777 /", ":(){ :|:& };:", "> /etc/",
        "del /f /q", "fdisk"
    ]

    # 正则匹配的危险模式：format 是 Windows 格式化磁盘命令（format C: /Q），
    # 但要避免误伤 PowerShell 的 "Get-Date -Format" 这类合法用法。
    DANGEROUS_REGEX = [
        re.compile(r'(^|[\s;&|])format\s+[a-zA-Z]:', re.IGNORECASE),
    ]

    # 自杀式进程杀检测：会杀掉 abyss 自身（python 进程）的命令。
    # 痛点：LLM 高频生成 `Stop-Process -Name python -Force` 来清理子进程，
    # 但 abyss 自己也是 python 进程，会被一起杀掉导致 CLI 退出。
    # 检测范围：Stop-Process -Name python* / taskkill /IM python.exe / killall python
    # 例外：用 -Id 精确指定 PID 的命令不拦截（合法用法）
    _SUICIDE_PATTERNS = [
        # PowerShell: Stop-Process ... -Name python/pythonw/python*（参数顺序无关）
        # 用 [^\n;]* 限制在同一语句内，避免跨语句误匹配
        re.compile(r'\bStop-Process\b[^\n;]*-Name\s+python\w*', re.IGNORECASE),
        # cmd: taskkill /F /IM python.exe / IM pythonw.exe
        re.compile(r'\btaskkill\b[^\n;]*\/IM\s+python\w*\.exe', re.IGNORECASE),
        # Unix: killall python / pkill -f python
        re.compile(r'\b(?:killall|pkill)\s+(?:-\w+\s+)*python\b', re.IGNORECASE),
    ]

    # 读取命令的正则模式：用于检测读敏感文件
    # 覆盖：cat / type / Get-Content / gc / Get-Item / gi
    # 注意：仅检测显式路径参数，不检测变量、管道输入
    _READ_CMD_PATTERNS = [
        # Unix: cat <path>  /  head -n N <path>  /  tail <path>  /  less <path>
        re.compile(r'\b(?:cat|head|tail|less|more)\s+["\']?([^\s"\';&|]+)["\']?'),
        # Windows cmd: type <path>
        re.compile(r'\btype\s+["\']?([^\s"\';&|]+)["\']?', re.IGNORECASE),
        # PowerShell: Get-Content <path> / gc <path>
        re.compile(r'\b(?:Get-Content|gc)\s+["\']?([^\s"\';&|]+)["\']?', re.IGNORECASE),
        # PowerShell: Get-Item <path> / gi <path>  （包含属性访问会泄露内容）
        re.compile(r'\b(?:Get-Item|gi)\s+["\']?([^\s"\';&|]+)["\']?', re.IGNORECASE),
    ]

    @property
    def name(self) -> str:
        return "shell_exec"

    @property
    def description(self) -> str:
        return "执行终端命令，支持设置工作目录和超时。失败时自动重试。"

    @property
    def parameters(self) -> dict:
        return {
            "type": "object",
            "properties": {
                "command": {"type": "string", "description": "要执行的命令"},
                "cwd": {"type": "string", "description": "工作目录"},
                "timeout": {"type": "integer", "description": "超时秒数，默认60"}
            },
            "required": ["command"]
        }

    def _is_safe(self, command: str) -> bool:
        """检查命令是否安全"""
        lowered = command.lower()
        for pattern in self.DANGEROUS_PATTERNS:
            if pattern in lowered:
                return False
        for pattern in self.DANGEROUS_REGEX:
            if pattern.search(command):
                return False
        return True

    @staticmethod
    def _detect_suicide(command: str) -> Optional[str]:
        """检测自杀式进程杀命令。命中返回原因，否则返回 None。

        痛点：LLM 生成 `Stop-Process -Name python -Force` 时会杀掉 abyss 自身
        （abyss 也是 python 进程），导致 CLI 退出。

        合法替代：用 `Start-Process ... -PassThru` 拿到进程对象，再
        `Stop-Process -Id $proc.Id` 精确杀指定 PID。
        """
        for pattern in ShellExecTool._SUICIDE_PATTERNS:
            if pattern.search(command):
                return (
                    "检测到自杀式进程杀命令（按名称杀 python 进程会同时杀掉 abyss 自身）。"
                    "请改用 -PassThru 拿到进程对象后按 -Id 精确杀指定 PID：\n"
                    "  $p = Start-Process python -ArgumentList xxx.py -PassThru\n"
                    "  Stop-Process -Id $p.Id -Force"
                )
        return None

    def _check_read_paths(self, command: str) -> Optional[str]:
        """检测读取命令的目标路径是否在工作目录外。

        防止 AI 通过 cat/Get-Content 等命令绕过 PathSecurityService 读取
        file_read 工具拒绝访问的敏感文件（如系统 hosts、其他项目源码）。

        Returns:
            拒绝原因字符串，None 表示放行。
        """
        from abyss.security.path_security import get_default_service

        svc = get_default_service()
        for pattern in self._READ_CMD_PATTERNS:
            for match in pattern.finditer(command):
                path = match.group(1)
                # 跳过 PowerShell 变量（$variable）和管道符
                if path.startswith("$") or path.startswith("|"):
                    continue
                # 跳过选项参数（-Force / -LiteralPath）
                if path.startswith("-"):
                    continue
                # 调用 PathSecurityService 校验
                check = svc.check_read(path)
                if not check.allowed and not check.needs_confirmation:
                    return f"shell_exec 读取路径被拒绝（{path}）：{check.reason}"
        return None

    @staticmethod
    def _enrich_timeout_error(command: str, timeout: int, result: dict) -> dict:
        """为超时错误附加针对性建议。目标：让 AI 看到错误后能立刻换策略。"""
        cmd_lower = command.lower()
        suggestions = []
        if re.search(r"\b(pip|pip3|python\s+-m\s+pip)\b", cmd_lower) and "install" in cmd_lower:
            suggestions.extend([
                "pip install 在国内常因 pypi 源慢而超时。",
                "建议加 -i https://pypi.tuna.tsinghua.edu.cn/simple（清华源）",
                "或阿里源 https://mirrors.aliyun.com/pypi/simple/",
                "并把 timeout 加大到 300+ 秒。",
            ])
        elif re.search(r"\bgit\s+clone\b", cmd_lower):
            suggestions.extend([
                "git clone 慢可换镜像：git clone https://mirror.ghproxy.com/<原github url>",
                "或 https://gitclone.com/<原 url>",
            ])
        elif re.search(r"\b(curl|wget|invoke-webrequest)\b", cmd_lower):
            suggestions.extend([
                "下载慢可换国内镜像或在浏览器中下载。",
            ])
        else:
            suggestions.append(
                f"已强制终止。建议把 timeout 调大（当前 {timeout}s），或拆成更小的步骤。"
            )
        if suggestions:
            new_err = result.get("error", "") + "\n[建议] " + " ".join(suggestions)
            result = dict(result)
            result["error"] = new_err
        return result

    def _wrap_powershell(self, command: str) -> str:
        """包装用户命令，强制 PowerShell 用 UTF-8 输出（避免中文乱码）。
        同时把 PYTHONIOENCODING=utf-8 写到子进程环境，让子 Python 进程也用 UTF-8 输出，
        避免 print('✓') 这类 Unicode 字符在 Windows GBK 终端上崩溃。

        关键：不能用 chcp 65001 或 [Console]::OutputEncoding = UTF8。
        两者都调用 SetConsoleOutputCP，跨进程改父控制台代码页，触发 conhost 清屏重绘
        —— 在 cmd.exe（默认 936）下表现为"第一个 shell 命令后前边记录消失"。
        改用 StreamWriter 包裹 stdout/stderr 管道写 UTF-8，完全不碰控制台 CP。
        """
        prefix = (
            "$OutputEncoding = [System.Text.Encoding]::UTF8; "
            "$u = New-Object Text.UTF8Encoding($false); "
            "$o = New-Object IO.StreamWriter([Console]::OpenStandardOutput(), $u); "
            "$o.AutoFlush = $true; [Console]::SetOut($o); "
            "$e = New-Object IO.StreamWriter([Console]::OpenStandardError(), $u); "
            "$e.AutoFlush = $true; [Console]::SetError($e); "
        )
        return prefix + command

    def _build_env(self) -> dict:
        """构造子进程环境：保留系统环境，并注入 PYTHONIOENCODING=utf-8"""
        import os
        env = os.environ.copy()
        env["PYTHONIOENCODING"] = "utf-8"
        env["PYTHONUTF8"] = "1"
        return env

    def _start_proc(self, command: str, cwd, env) -> subprocess.Popen:
        """启动子进程并返回 Popen 对象。Windows 走 PowerShell，其他平台走 sh。"""
        if sys.platform == "win32":
            return subprocess.Popen(
                ["powershell.exe", "-NoProfile", "-Command", self._wrap_powershell(command)],
                cwd=cwd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                encoding="utf-8",
                errors="replace",
                env=env,
            )
        return subprocess.Popen(
            command,
            shell=True,
            cwd=cwd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
            errors="replace",
            env=env,
        )

    @staticmethod
    def _start_reader(stream, sink: list) -> threading.Thread:
        """启动后台线程，把 stream 读到 sink 列表中。读到 EOF 自然退出。

        用底层 buffer.read1() 而非 TextIOWrapper.read()：read() 会试图填满
        内部缓冲区，孙进程持有管道时它会阻塞在「再读一点」上，导致已缓冲的
        PassThru 输出也读不出来。read1() 有多少读多少，立刻返回。
        """
        def _read_stream():
            try:
                raw = getattr(stream, "buffer", stream)
                while True:
                    chunk = raw.read1(4096) if hasattr(raw, "read1") else raw.read(4096)
                    if not chunk:
                        break
                    if isinstance(chunk, bytes):
                        sink.append(chunk.decode("utf-8", errors="replace"))
                    else:
                        sink.append(chunk)
            except Exception:
                pass
        thread = threading.Thread(target=_read_stream, daemon=True)
        thread.start()
        return thread

    @staticmethod
    def _join_readers(*threads, drain_timeout: float = 2.0):
        """等待读取线程结束，最多 drain_timeout 秒。超时后线程留在后台。

        关键：孙进程可能继承管道写端（Start-Process -NoNewWindow 场景），
        此时读取线程永远读不到 EOF。我们不能等它 — 直接返回已读到的内容，
        让孙进程在后台继续运行（这正是用户启动后台服务的本意）。
        """
        deadline = time.monotonic() + drain_timeout
        for thread in threads:
            remaining = max(0.0, deadline - time.monotonic())
            thread.join(timeout=remaining)

    @staticmethod
    def _kill_tree(pid: int):
        """杀掉进程及其所有子孙进程。Windows 用 taskkill /T，POSIX 用进程组。"""
        if sys.platform == "win32":
            subprocess.run(
                f"taskkill /F /T /PID {pid}",
                shell=True,
                capture_output=True,
            )
        else:
            try:
                os.killpg(os.getpgid(pid), signal.SIGKILL)
            except (ProcessLookupError, PermissionError):
                pass

    def _run_once(self, command: str, cwd, timeout) -> dict:
        """单次执行命令（不重试）。

        用 Popen + 后台线程读取，而不是 subprocess.run(capture_output=True)。
        原因：subprocess.run 内部调用 communicate() 等待 stdout/stderr 管道 EOF，
        当用户执行 `Start-Process -NoNewWindow -PassThru <长进程>` 时，
        孙进程会继承 PowerShell 的管道写端，PowerShell 退出后管道仍被持有，
        communicate() 永远读不到 EOF → 工具卡死。

        新方案：用 proc.wait() 只等 PowerShell 进程退出（与管道 EOF 解耦），
        再用读取线程带超时地拉取已缓冲的输出。
        """
        env = self._build_env()
        proc = self._start_proc(command, cwd, env)
        stdout_chunks: list = []
        stderr_chunks: list = []
        stdout_thread = self._start_reader(proc.stdout, stdout_chunks)
        stderr_thread = self._start_reader(proc.stderr, stderr_chunks)

        try:
            proc.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            # 父进程仍在运行 — 真正的超时，杀整棵进程树（含孙进程）
            self._kill_tree(proc.pid)
            try:
                proc.wait(timeout=5)
            except subprocess.TimeoutExpired:
                pass
            self._join_readers(stdout_thread, stderr_thread, drain_timeout=1.0)
            return {
                "command": command,
                "success": False,
                "returncode": proc.returncode,
                "stdout": "".join(stdout_chunks),
                "stderr": "".join(stderr_chunks),
                "error": f"命令执行超时（{timeout}秒），已强制终止",
            }

        # 父进程已退出 — 给读取线程一小段时间把缓冲读干。
        # 若孙进程仍持有管道（Start-Process 场景），线程不会结束，
        # 此时直接返回已读内容，让孙进程继续在后台运行。
        self._join_readers(stdout_thread, stderr_thread, drain_timeout=2.0)
        return {
            "command": command,
            "success": proc.returncode == 0,
            "returncode": proc.returncode,
            "stdout": "".join(stdout_chunks),
            "stderr": "".join(stderr_chunks),
        }

    def execute(self, command: str, cwd: str = None, timeout: int = 60) -> dict:
        """执行命令，失败时按指数退避自动重试。超时场景不重试（重试也没用）"""
        # 兼容 LLM 将 timeout 序列化为字符串的情形
        timeout = coerce_int(timeout, default=60)

        # 同步 LLM 传入的 cwd 到 PathSecurityService（修复 file_write 边界不一致）
        # 场景：用户在 A 目录启动 abyss，/quit 后在 B 目录重启并 /session 恢复会话。
        # AI 从历史里"记住"原工作目录 A，通过 shell_exec(cwd=A) 执行命令成功，
        # 但 file_write 无 cwd 参数，PathSecurityService 仍用 os.getcwd()=B 做边界，
        # 导致写入 A 目录下的文件被误判为 "path outside workspace boundary"。
        # 这里让 file_write/file_read 的边界跟随 LLM 最近一次 shell_exec 的 cwd。
        if cwd:
            from abyss.security.path_security import get_default_service as _get_path_sec
            _get_path_sec().set_working_dir(cwd)

        if not self._is_safe(command):
            return {"success": False, "error": "命令被安全策略拦截"}

        # 自杀式进程杀检测：拦截 Stop-Process -Name python 等会杀掉 abyss 自身的命令
        suicide_reason = self._detect_suicide(command)
        if suicide_reason is not None:
            return {"success": False, "error": suicide_reason}

        # ToolSafetyGuard 多层纵深防护（Phase 8 CE）
        # 与 _is_safe 互补：_is_safe 偏 Windows（format C:），
        # ToolSafetyGuard 偏 Unix（fork bomb / mkfs / curl|sh / shutdown 等 12 类）
        from ..security.tool_safety_guard import ToolSafetyGuard
        guard = ToolSafetyGuard(working_dir=cwd or os.getcwd())
        guard_result = guard.check_bash_command(command)
        if guard_result.should_block:
            return {
                "success": False,
                "error": f"命令被 ToolSafetyGuard 拦截：{guard_result.reason}",
            }

        # 路径安全校验：检测 cat/type/Get-Content 等读取命令是否在工作目录外
        # 防止 AI 绕过 file_read 工具的安全防护读取敏感文件
        read_block_reason = self._check_read_paths(command)
        if read_block_reason is not None:
            return {"success": False, "error": read_block_reason}

        last_result = None
        for attempt in range(self.MAX_RETRIES + 1):
            try:
                last_result = self._run_once(command, cwd, timeout)
            except subprocess.TimeoutExpired:
                # 超时不重试：第一次就附加建议，避免无意义的重试浪费时间
                last_result = {
                    "command": command,
                    "success": False,
                    "error": f"命令执行超时（{timeout}秒），已强制终止",
                }
                return self._enrich_timeout_error(command, timeout, last_result)
            except Exception as e:
                last_result = {"command": command, "success": False, "error": str(e)}

            # 如果是超时（不抛异常时由 _run_once 返回 success=False 且 error 含"超时"），
            # 不再重试，附加建议后直接返回。
            err_msg = (last_result or {}).get("error", "") or ""
            if not last_result.get("success") and "超时" in err_msg:
                return self._enrich_timeout_error(command, timeout, last_result)

            if last_result.get("success"):
                return last_result

            # 失败：除最后一次外都退避
            if attempt < self.MAX_RETRIES:
                time.sleep(self.BASE_BACKOFF * (2 ** attempt))

        return last_result



# -*- coding: utf-8 -*-
"""
工具模块
提供文件操作、命令执行、联网搜索等工具
"""
from .base import Tool
from .file_read import FileReadTool
from .file_write import FileWriteTool
from .file_edit import FileEditTool
from .shell_exec import ShellExecTool
from .web_search import WebSearchTool
from .web_fetch import WebFetchTool

__all__ = [
    "Tool",
    "FileReadTool",
    "FileWriteTool",
    "FileEditTool",
    "ShellExecTool",
    "WebSearchTool",
    "WebFetchTool",
]



# -*- coding: utf-8 -*-
"""
回归测试：工具参数归一化守卫。

针对 LLM 偶尔把 list/dict/int/bool 参数序列化为 JSON 字符串的场景，
coerce 系列函数应安全还原为原生类型，避免下游迭代字符串字符或类型错误。
"""
import pytest

from abyss.tools.base import (
    coerce_int,
    coerce_bool,
    coerce_list_of_dicts,
    coerce_list_of_str,
)


# ---------- coerce_int ----------

class TestCoerceInt:
    """验证 int 参数归一化。"""

    @pytest.mark.parametrize("value, expected", [
        (10, 10),
        ("10", 10),
        ("0", 0),
        ("-5", -5),
        (10.0, 10),
        ("10.5", 10),  # float 字符串截断
    ])
    def test_valid_inputs(self, value, expected):
        assert coerce_int(value, default=5) == expected

    @pytest.mark.parametrize("value", [
        None,
        "",
        "abc",
        [],
        {},
        "1.2.3",
    ])
    def test_invalid_returns_default(self, value):
        assert coerce_int(value, default=42) == 42

    def test_none_with_no_default_returns_zero(self):
        assert coerce_int(None) == 0


# ---------- coerce_bool ----------

class TestCoerceBool:
    """验证 bool 参数归一化，特别防御字符串 'false' 被判 truthy。"""

    @pytest.mark.parametrize("value, expected", [
        (True, True),
        (False, False),
        ("true", True),
        ("false", False),
        ("True", True),
        ("False", False),
        ("TRUE", True),
        ("FALSE", False),
        (1, True),
        (0, False),
        ("1", True),
        ("0", False),
        ("yes", True),
        ("no", False),
        ("on", True),
        ("off", False),
    ])
    def test_truthy_falsy_strings(self, value, expected):
        assert coerce_bool(value) is expected

    @pytest.mark.parametrize("value", [
        None,
        "",
        [],
        {},
        "abc",
    ])
    def test_invalid_returns_default(self, value):
        assert coerce_bool(value, default=False) is False
        assert coerce_bool(value, default=True) is True


# ---------- coerce_list_of_dicts ----------

class TestCoerceListOfDicts:
    """验证 list[dict] 参数归一化（todo_write / spawn_subagents 场景）。"""

    def test_native_list_of_dicts(self):
        todos = [{"id": "1"}, {"id": "2"}]
        assert coerce_list_of_dicts(todos) == todos

    def test_json_string_of_list(self):
        s = '[{"id": "1"}, {"id": "2"}]'
        assert coerce_list_of_dicts(s) == [{"id": "1"}, {"id": "2"}]

    def test_json_string_filters_non_dict_items(self):
        s = '[{"id": "1"}, "junk", 42, null]'
        assert coerce_list_of_dicts(s) == [{"id": "1"}]

    def test_invalid_json_string_returns_empty(self):
        assert coerce_list_of_dicts("not json") == []

    @pytest.mark.parametrize("value", [
        None,
        "",
        [],
        {},
        42,
        True,
    ])
    def test_non_list_returns_empty(self, value):
        assert coerce_list_of_dicts(value) == []

    def test_filters_non_dict_elements_in_native_list(self):
        items = [{"ok": 1}, "bad", 42, None, {"ok": 2}]
        assert coerce_list_of_dicts(items) == [{"ok": 1}, {"ok": 2}]


# ---------- coerce_list_of_str ----------

class TestCoerceListOfStr:
    """验证 list[str] 参数归一化（knowledge_save tags 场景）。"""

    def test_native_list_of_str(self):
        tags = ["a", "b"]
        assert coerce_list_of_str(tags) == ["a", "b"]

    def test_json_string_of_list(self):
        s = '["a", "b"]'
        assert coerce_list_of_str(s) == ["a", "b"]

    def test_json_string_filters_non_str_items(self):
        s = '["a", 42, null, {"x": 1}]'
        assert coerce_list_of_str(s) == ["a"]

    def test_invalid_json_returns_empty(self):
        assert coerce_list_of_str("not json") == []

    @pytest.mark.parametrize("value", [
        None,
        "",
        [],
        {},
        42,
    ])
    def test_non_list_returns_empty(self, value):
        assert coerce_list_of_str(value) == []

    def test_filters_non_str_elements_in_native_list(self):
        items = ["a", 42, None, {"x": 1}, "b"]
        assert coerce_list_of_str(items) == ["a", "b"]



# -*- coding: utf-8 -*-
"""CN InfoHelpCommands 测试（Phase 8 Tier 2）

借鉴 zhikuncode InfoHelpCommands.java：
- /stats — 会话统计（消息数/工具调用数/文件修改数/持续时间）
- /advisor — AI 顾问模式（不直接改代码，只给建议）
- /btw — 顺便提问（不中断当前任务）
- /version — 版本信息（Python/abyss/OS）
"""
import platform
import sys
from unittest.mock import MagicMock

import pytest

from abyss.commands.info import (
    handle_stats,
    handle_advisor,
    handle_btw,
    handle_version,
    count_messages_by_role,
    count_tool_calls,
    format_duration,
)


def _make_messages_with_tool_calls() -> list:
    """构造含 tool_calls 的消息"""
    return [
        {"role": "system", "content": "sys"},
        {"role": "user", "content": "edit demo.py"},
        {
            "role": "assistant",
            "content": None,
            "tool_calls": [
                {"id": "1", "type": "function", "function": {"name": "file_edit", "arguments": "{}"}},
                {"id": "2", "type": "function", "function": {"name": "file_read", "arguments": "{}"}},
            ],
        },
        {"role": "tool", "content": "result 1", "tool_call_id": "1"},
        {"role": "tool", "content": "result 2", "tool_call_id": "2"},
        {"role": "assistant", "content": "done"},
    ]


class TestCountMessagesByRole:
    """count_messages_by_role 测试"""

    def test_counts_each_role(self):
        messages = _make_messages_with_tool_calls()
        counts = count_messages_by_role(messages)
        assert counts["system"] == 1
        assert counts["user"] == 1
        assert counts["assistant"] == 2
        assert counts["tool"] == 2

    def test_empty_messages(self):
        assert count_messages_by_role([]) == {}

    def test_unknown_role_counted(self):
        messages = [{"role": "developer", "content": "x"}]
        counts = count_messages_by_role(messages)
        assert counts["developer"] == 1


class TestCountToolCalls:
    """count_tool_calls 测试"""

    def test_counts_tool_calls_in_assistant_messages(self):
        messages = _make_messages_with_tool_calls()
        assert count_tool_calls(messages) == 2

    def test_zero_when_no_tool_calls(self):
        messages = [{"role": "user", "content": "hi"}]
        assert count_tool_calls(messages) == 0

    def test_empty_messages(self):
        assert count_tool_calls([]) == 0


class TestFormatDuration:
    """format_duration 测试"""

    def test_seconds(self):
        assert format_duration(45) == "45s"

    def test_minutes(self):
        assert format_duration(120) == "2m 0s"

    def test_hours(self):
        assert format_duration(3661) == "1h 1m 1s"

    def test_zero(self):
        assert format_duration(0) == "0s"


class TestHandleStats:
    """handle_stats 测试"""

    def test_counts_messages_correctly(self):
        session = MagicMock()
        session.messages = _make_messages_with_tool_calls()
        result = handle_stats(session)
        assert "user" in result.lower() or "用户" in result
        assert "assistant" in result.lower() or "助手" in result or "AI" in result
        assert "tool" in result.lower() or "工具" in result

    def test_counts_tool_calls(self):
        session = MagicMock()
        session.messages = _make_messages_with_tool_calls()
        result = handle_stats(session)
        assert "tool_call" in result.lower() or "工具调用" in result or "2" in result

    def test_shows_session_duration(self):
        session = MagicMock()
        session.messages = _make_messages_with_tool_calls()
        session.started_at = None
        result = handle_stats(session)
        # started_at None 时优雅降级
        assert isinstance(result, str)

    def test_empty_session_handled(self):
        session = MagicMock()
        session.messages = []
        result = handle_stats(session)
        assert isinstance(result, str)


class TestHandleAdvisor:
    """handle_advisor 测试"""

    def test_injects_advisory_prompt(self):
        session = MagicMock()
        result = handle_advisor(session, "how to improve my code")
        assert result == "injected"
        session.add_user.assert_called_once()
        args = session.add_user.call_args[0]
        assert "顾问" in args[0] or "advisor" in args[0].lower()
        assert "how to improve my code" in args[0]

    def test_no_target_returns_error(self):
        session = MagicMock()
        result = handle_advisor(session, "")
        assert result != "injected"
        assert "用法" in result or "usage" in result.lower()
        session.add_user.assert_not_called()


class TestHandleBtw:
    """handle_btw 测试"""

    def test_injects_inline_question(self):
        session = MagicMock()
        result = handle_btw(session, "what does async mean")
        assert result == "injected"
        session.add_user.assert_called_once()
        args = session.add_user.call_args[0]
        assert "顺便" in args[0] or "btw" in args[0].lower() or "interrupt" in args[0].lower()
        assert "what does async mean" in args[0]

    def test_no_target_returns_error(self):
        session = MagicMock()
        result = handle_btw(session, "")
        assert result != "injected"
        session.add_user.assert_not_called()


class TestHandleVersion:
    """handle_version 测试"""

    def test_shows_python_version(self):
        result = handle_version()
        assert "Python" in result
        assert platform.python_version() in result or "3." in result

    def test_shows_abyss_version(self):
        result = handle_version()
        # abyss version 来自 __init__.py
        assert "abyss" in result.lower() or "Abyss" in result

    def test_shows_os_info(self):
        result = handle_version()
        assert platform.system() in result or "Windows" in result or "Linux" in result or "Darwin" in result

    def test_returns_string(self):
        result = handle_version()
        assert isinstance(result, str)
        assert len(result) > 0



# -*- coding: utf-8 -*-
"""CJ DiffCommand 测试（Phase 8 Tier 2）

借鉴 zhikuncode DiffCommand.java：
- /diff [staged] 显示 Git 工作区或暂存区差异
- 检测当前目录是否 git 仓库
- 截断 10000 字符避免刷屏
- 同时返回 stat 摘要
"""
import subprocess
from pathlib import Path
from unittest.mock import patch, MagicMock

import pytest

from abyss.commands.diff import (
    handle_diff,
    is_git_repo,
    MAX_DIFF_OUTPUT_CHARS,
    run_git_diff,
    run_git_stat,
)


class TestIsGitRepo:
    """is_git_repo 测试"""

    def test_returns_true_in_git_repo(self, tmp_path):
        """在 git 仓库内返回 True"""
        # init a git repo
        subprocess.run(["git", "init"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.email", "test@test.com"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.name", "test"], cwd=str(tmp_path), capture_output=True)
        assert is_git_repo(str(tmp_path)) is True

    def test_returns_false_in_non_git_dir(self, tmp_path):
        """非 git 目录返回 False"""
        assert is_git_repo(str(tmp_path)) is False

    def test_returns_false_for_nonexistent_dir(self):
        """不存在的目录返回 False"""
        assert is_git_repo("Z:/nonexistent/path/xyz") is False


class TestRunGitDiff:
    """run_git_diff 测试"""

    def test_unstaged_calls_git_diff(self, tmp_path):
        """unstaged 调用 git diff"""
        subprocess.run(["git", "init"], cwd=str(tmp_path), capture_output=True)
        # 创建文件并 commit
        (tmp_path / "a.txt").write_text("initial", encoding="utf-8")
        subprocess.run(["git", "add", "."], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "commit", "-m", "init"], cwd=str(tmp_path), capture_output=True)
        # 修改文件
        (tmp_path / "a.txt").write_text("modified", encoding="utf-8")
        out, err = run_git_diff(str(tmp_path), staged=False)
        assert "modified" in out or "initial" in out

    def test_staged_calls_git_diff_cached(self, tmp_path):
        """staged 调用 git diff --cached"""
        subprocess.run(["git", "init"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.email", "t@t.com"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.name", "t"], cwd=str(tmp_path), capture_output=True)
        (tmp_path / "a.txt").write_text("initial", encoding="utf-8")
        subprocess.run(["git", "add", "."], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "commit", "-m", "init"], cwd=str(tmp_path), capture_output=True)
        # 修改并 stage
        (tmp_path / "a.txt").write_text("staged modification", encoding="utf-8")
        subprocess.run(["git", "add", "a.txt"], cwd=str(tmp_path), capture_output=True)
        out, err = run_git_diff(str(tmp_path), staged=True)
        assert "staged modification" in out


class TestRunGitStat:
    """run_git_stat 测试"""

    def test_returns_stat_output(self, tmp_path):
        """返回 stat 摘要"""
        subprocess.run(["git", "init"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.email", "t@t.com"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.name", "t"], cwd=str(tmp_path), capture_output=True)
        (tmp_path / "a.txt").write_text("initial", encoding="utf-8")
        subprocess.run(["git", "add", "."], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "commit", "-m", "init"], cwd=str(tmp_path), capture_output=True)
        (tmp_path / "a.txt").write_text("modified content", encoding="utf-8")
        out, _ = run_git_stat(str(tmp_path))
        assert "a.txt" in out


class TestHandleDiff:
    """handle_diff 测试"""

    def test_not_git_repo_returns_error(self, tmp_path):
        """非 git 仓库返回错误消息"""
        result = handle_diff(str(tmp_path), staged=False)
        assert "不是 Git 仓库" in result or "not a git" in result.lower()

    def test_returns_diff_output_for_unstaged(self, tmp_path):
        """unstaged 返回 diff"""
        subprocess.run(["git", "init"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.email", "t@t.com"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.name", "t"], cwd=str(tmp_path), capture_output=True)
        (tmp_path / "a.txt").write_text("initial", encoding="utf-8")
        subprocess.run(["git", "add", "."], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "commit", "-m", "init"], cwd=str(tmp_path), capture_output=True)
        (tmp_path / "a.txt").write_text("modified content", encoding="utf-8")
        result = handle_diff(str(tmp_path), staged=False)
        assert "modified content" in result or "initial" in result

    def test_returns_diff_output_for_staged(self, tmp_path):
        """staged 返回 cached diff"""
        subprocess.run(["git", "init"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.email", "t@t.com"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.name", "t"], cwd=str(tmp_path), capture_output=True)
        (tmp_path / "a.txt").write_text("initial", encoding="utf-8")
        subprocess.run(["git", "add", "."], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "commit", "-m", "init"], cwd=str(tmp_path), capture_output=True)
        (tmp_path / "a.txt").write_text("staged content", encoding="utf-8")
        subprocess.run(["git", "add", "a.txt"], cwd=str(tmp_path), capture_output=True)
        result = handle_diff(str(tmp_path), staged=True)
        assert "staged content" in result

    def test_output_truncated_at_max(self, tmp_path):
        """超过 MAX_DIFF_OUTPUT_CHARS 时截断"""
        subprocess.run(["git", "init"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.email", "t@t.com"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.name", "t"], cwd=str(tmp_path), capture_output=True)
        (tmp_path / "a.txt").write_text("initial", encoding="utf-8")
        subprocess.run(["git", "add", "."], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "commit", "-m", "init"], cwd=str(tmp_path), capture_output=True)
        # 构造超长 diff
        (tmp_path / "a.txt").write_text("x" * (MAX_DIFF_OUTPUT_CHARS + 1000), encoding="utf-8")
        result = handle_diff(str(tmp_path), staged=False)
        assert "截断" in result or "truncated" in result.lower()

    def test_includes_stat_summary(self, tmp_path):
        """结果包含 stat 摘要"""
        subprocess.run(["git", "init"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.email", "t@t.com"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.name", "t"], cwd=str(tmp_path), capture_output=True)
        (tmp_path / "a.txt").write_text("initial", encoding="utf-8")
        subprocess.run(["git", "add", "."], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "commit", "-m", "init"], cwd=str(tmp_path), capture_output=True)
        (tmp_path / "a.txt").write_text("modified content here", encoding="utf-8")
        result = handle_diff(str(tmp_path), staged=False)
        assert "a.txt" in result  # stat 中应包含文件名

    def test_no_changes_returns_info_message(self, tmp_path):
        """无变更时返回提示消息"""
        subprocess.run(["git", "init"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.email", "t@t.com"], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "config", "user.name", "t"], cwd=str(tmp_path), capture_output=True)
        (tmp_path / "a.txt").write_text("initial", encoding="utf-8")
        subprocess.run(["git", "add", "."], cwd=str(tmp_path), capture_output=True)
        subprocess.run(["git", "commit", "-m", "init"], cwd=str(tmp_path), capture_output=True)
        result = handle_diff(str(tmp_path), staged=False)
        assert "无变更" in result or "no changes" in result.lower()
